import copy
import numpy as np
import itertools
import unicodedata
import re
from fantasybert.tokenization.tokenization_utils import to_py_obj
from typing import Dict, List, Optional, Tuple, Union
from fantasybert.tokenization.tokenization_utils import PaddingStrategy, TensorType, TruncationStrategy, _is_control,\
    SpecialTokensMixin, BatchEncoding, AddedToken, _insert_one_token_to_ordered_list, _is_end_of_word, _is_start_of_word

TextInput = str
PreTokenizedInput = List[str]
EncodedInput = List[int]
TextInputPair = Tuple[str, str]
PreTokenizedInputPair = Tuple[List[str], List[str]]
EncodedInputPair = Tuple[List[int], List[int]]

VERY_LARGE_INTEGER = int(1e30)  # This is used to set the max input length for a model with infinite size input
LARGE_INTEGER = int(1e20)  # This is used when we need something big but slightly smaller than VERY_LARGE_INTEGER


class TokenizerBase(SpecialTokensMixin):
    def __init__(self, **kwargs):
        self.init_inputs = ()
        self.init_kwargs = copy.deepcopy(kwargs)

        self.added_tokens_encoder: Dict[str, int] = {}
        self.added_tokens_decoder: Dict[int, str] = {}
        self.unique_no_split_tokens: List[str] = []

        model_max_length = kwargs.pop("model_max_length", kwargs.pop("max_len", None))
        self.model_max_length = model_max_length if model_max_length is not None else VERY_LARGE_INTEGER
        super().__init__(**kwargs)

    def _add_tokens(self, new_tokens: Union[List[str], List[AddedToken]], special_tokens: bool = False) -> int:
        new_tokens = [str(tok) for tok in new_tokens]

        tokens_to_add = []
        for token in new_tokens:
            assert isinstance(token, str)
            if not special_tokens and hasattr(self, "do_lower_case") and self.do_lower_case:
                token = token.lower()
            if (
                token != self.unk_token
                and self.convert_tokens_to_ids(token) == self.convert_tokens_to_ids(self.unk_token)
                and token not in tokens_to_add
            ):
                tokens_to_add.append(token)


        added_tok_encoder = dict((tok, len(self) + i) for i, tok in enumerate(tokens_to_add))
        added_tok_decoder = {v: k for k, v in added_tok_encoder.items()}
        self.added_tokens_encoder.update(added_tok_encoder)
        self.added_tokens_decoder.update(added_tok_decoder)

        # Make sure we don't split on any special tokens (even they were already in the vocab before e.g. for Albert)
        if special_tokens:
            if len(new_tokens) == 1:
                _insert_one_token_to_ordered_list(self.unique_no_split_tokens, new_tokens[0])
            else:
                self.unique_no_split_tokens = sorted(set(self.unique_no_split_tokens).union(set(new_tokens)))
        else:              # Or on the newly added tokens
            if len(tokens_to_add) == 1:
                _insert_one_token_to_ordered_list(self.unique_no_split_tokens, tokens_to_add[0])
            else:
                self.unique_no_split_tokens = sorted(set(self.unique_no_split_tokens).union(set(tokens_to_add)))

        return len(tokens_to_add)

    @property
    def vocab_size(self) -> int:
        raise NotImplementedError

    @property
    def do_lower_case(self) -> bool:
        raise NotImplementedError

    def __len__(self):
        return self.vocab_size + len(self.added_tokens_encoder)

    def get_vocab(self) -> Dict[str, int]:
        raise NotImplementedError()

    def tokenize(self, text: str, pair: Optional[str] = None, add_special_tokens: bool = False, **kwargs) -> List[str]:
        all_special_tokens_extended = dict((str(t), t) for t in self.all_special_tokens if isinstance(t, AddedToken))

        if hasattr(self, "do_lower_case") and self.do_lower_case:
            # convert non-special tokens to lowercase
            escaped_special_toks = [re.escape(s_tok) for s_tok in self.all_special_tokens]
            pattern = r"(" + r"|".join(escaped_special_toks) + r")|" + r"(.+?)"
            text = re.sub(pattern, lambda m: m.groups()[0] or m.groups()[1].lower(), text)

        def split_on_token(tok, text):
            result = []
            tok_extended = all_special_tokens_extended.get(tok, None)
            split_text = text.split(tok)
            full_word = ""
            for i, sub_text in enumerate(split_text):
                if isinstance(tok_extended, AddedToken):
                    if tok_extended.single_word:
                        # Try to avoid splitting on token
                        if (
                                i < len(split_text) - 1
                                and not _is_end_of_word(sub_text)
                                and not _is_start_of_word(split_text[i + 1])
                        ):
                            # Don't extract the special token
                            full_word += sub_text + tok
                        elif full_word:
                            full_word += sub_text
                            result.append(full_word)
                            full_word = ""
                            continue
                    # Strip white spaces on the right
                    if tok_extended.rstrip and i > 0:
                        # A bit counter-intuitive but we strip the left of the string
                        # since tok_extended.rstrip means the special token is eating all white spaces on its right
                        sub_text = sub_text.lstrip()
                    # Strip white spaces on the left
                    if tok_extended.lstrip and i < len(split_text) - 1:
                        sub_text = sub_text.rstrip()  # Opposite here
                else:
                    # We strip left and right by default
                    if i < len(split_text) - 1:
                        sub_text = sub_text.rstrip()
                    if i > 0:
                        sub_text = sub_text.lstrip()

                if i == 0 and not sub_text:
                    result.append(tok)
                elif i == len(split_text) - 1:
                    if sub_text:
                        result.append(sub_text)
                    else:
                        pass
                else:
                    if sub_text:
                        result.append(sub_text)
                    result.append(tok)
            return result

        def split_on_tokens(tok_list, text):
            if not text.strip():
                return []
            if not tok_list:
                return self._tokenize(text)

            tokenized_text = []
            text_list = [text]
            for tok in tok_list:
                tokenized_text = []
                for sub_text in text_list:
                    if sub_text not in self.unique_no_split_tokens:
                        tokenized_text.extend(split_on_token(tok, sub_text))
                    else:
                        tokenized_text.append(sub_text)
                text_list = tokenized_text

            return list(
                itertools.chain.from_iterable(
                    (
                        self._tokenize(token) if token not in self.unique_no_split_tokens else [token]
                        for token in tokenized_text
                    )
                )
            )

        no_split_token = self.unique_no_split_tokens
        tokenized_text = split_on_tokens(no_split_token, text)
        return tokenized_text

    def _tokenize(self, text):
        raise NotImplementedError

    def _convert_token_to_id_with_added_voc(self, token):
        if token is None:
            return None

        if token in self.added_tokens_encoder:
            return self.added_tokens_encoder[token]
        return self._convert_token_to_id(token)

    def convert_tokens_to_ids(self, tokens: Union[str, List[str]]) -> Union[int, List[int]]:
        if isinstance(tokens, str):
            return self._convert_token_to_id_with_added_voc(tokens)
        ids = []
        for token in tokens:
            ids.append(self._convert_token_to_id_with_added_voc(token))
        return ids

    def convert_ids_to_tokens(self, ids: Union[int, List[int]], skip_special_tokens: bool = False) -> Union[str, List[str]]:

        if isinstance(ids, int):
            if ids in self.added_tokens_decoder:
                return self.added_tokens_decoder[ids]
            else:
                return self._convert_id_to_token(ids)
        tokens = []
        for index in ids:
            index = int(index)
            if skip_special_tokens and index in self.all_special_ids:
                continue
            if index in self.added_tokens_decoder:
                tokens.append(self.added_tokens_decoder[index])
            else:
                tokens.append(self._convert_id_to_token(index))
        return tokens

    def _convert_token_to_id(self, token):
        raise NotImplementedError

    def _convert_id_to_token(self, index: int) -> str:
        raise NotImplementedError

    def convert_tokens_to_string(self, tokens: List[str]) -> str:
        return " ".join(tokens)

    def encode(
        self,
        text: Union[TextInput, PreTokenizedInput, EncodedInput],
        text_pair: Optional[Union[TextInput, PreTokenizedInput, EncodedInput]] = None,
        add_special_tokens: bool = True,
        padding: Union[bool, str, PaddingStrategy] = False,
        truncation: Union[bool, str, TruncationStrategy] = False,
        max_length: Optional[int] = None,
        return_tensors: Optional[Union[str, TensorType]] = None,
        **kwargs
    ) -> List[int]:

        encoded_inputs = self.encode_plus(
            text,
            text_pair=text_pair,
            add_special_tokens=add_special_tokens,
            padding=padding,
            truncation=truncation,
            max_length=max_length,
            return_tensors=return_tensors,
            **kwargs)
        return encoded_inputs["input_ids"]

    def num_special_tokens_to_add(self, pair: bool = False) -> int:
        # 统计token个数
        token_ids_0 = []
        token_ids_1 = []
        return len(self.build_inputs_with_special_tokens(token_ids_0, token_ids_1 if pair else None))

    def _get_padding_truncation_strategies(
        self, padding=False, truncation=False, max_length=None, **kwargs
    ):
        """
        获取 padding 和 truncation的方式
        """
        padding_strategy, truncation_strategy = None, None

        if max_length is not None and padding is False and truncation is False:
            # 未给定最大句长和不进行padding时，则必须要进行截断
            truncation_strategy = (TruncationStrategy.LONGEST_FIRST)  # 当输入多个句子时，对最长的句子进行截断

        # Get padding strategy
        if padding is False:
            if max_length is None:
                padding_strategy = PaddingStrategy.LONGEST  # 若未给出最大句长，则默认根据batch中最长句子长度进行padding
            else:
                padding_strategy = PaddingStrategy.MAX_LENGTH
        elif padding is not False:
            if padding is True:
                padding_strategy = PaddingStrategy.LONGEST
            elif not isinstance(padding, PaddingStrategy):  # padding不是PaddingStrategy类型,则为str，传入PaddingStrategy
                padding_strategy = PaddingStrategy(padding)
            elif isinstance(padding, PaddingStrategy):
                padding_strategy = padding
        else:
            padding_strategy = PaddingStrategy.DO_NOT_PAD

        # Get truncation strategy
        if truncation is not False:
            if truncation is True:
                truncation_strategy = (
                    TruncationStrategy.LONGEST_FIRST
                )
            elif not isinstance(truncation, TruncationStrategy):
                truncation_strategy = TruncationStrategy(truncation)
            elif isinstance(truncation, TruncationStrategy):
                truncation_strategy = truncation
        else:
            truncation_strategy = TruncationStrategy.DO_NOT_TRUNCATE

        # Set max length if needed
        if max_length is None:
            if padding_strategy == PaddingStrategy.MAX_LENGTH:
                if self.model_max_length > LARGE_INTEGER:

                    padding_strategy = PaddingStrategy.DO_NOT_PAD
                else:
                    max_length = self.model_max_length

            if truncation_strategy != TruncationStrategy.DO_NOT_TRUNCATE:
                if self.model_max_length > LARGE_INTEGER:
                    truncation_strategy = TruncationStrategy.DO_NOT_TRUNCATE
                else:
                    max_length = self.model_max_length
        return padding_strategy, truncation_strategy, max_length, kwargs

    def __call__(
        self,
        text: Union[TextInput, PreTokenizedInput, List[TextInput], List[PreTokenizedInput]],
        text_pair: Optional[Union[TextInput, PreTokenizedInput, List[TextInput], List[PreTokenizedInput]]] = None,
        add_special_tokens: bool = True,  # 是否添加[CLS],[SEP]等特殊token
        padding: Union[bool, str, PaddingStrategy] = False,  # 是否进行填充 可填入的值为True/False/longest/batch_longest/max_length
        truncation: Union[bool, str, TruncationStrategy] = False,  # 是否进行截断 可填入的值为True/False/only_first/only_second/longest_first
        max_length: Optional[int] = None,  # 句子的最大长度，用于短填长切
        stride: int = 0,  #截断时使用的滑窗
        is_split_into_words: bool = False,  #  输入是否已经分好词了
        return_tensors: Optional[Union[str, TensorType]] = None,  # True返回tensor else List
        return_overflowing_tokens: bool = False,   # 是否返回截断超出的tokens
        return_special_tokens_mask: bool = False,  # 针对特殊token的mask，对应位置为1
        return_offsets_mapping: bool = False,    # 是否返回 word位置:token位置  (英文一个word可能会被拆分为几个token(subword))
        **kwargs) -> BatchEncoding:

        # 检查输入格式错误
        assert isinstance(text, str) or (
            isinstance(text, (list, tuple))
            and (
                len(text) == 0
                or (
                    isinstance(text[0], str)
                    or (isinstance(text[0], (list, tuple)) and (len(text[0]) == 0 or isinstance(text[0][0], str)))
                )
            )
        ), (
            "text input must of type `str` (single example), `List[str]` (batch or single pretokenized example) "
            "or `List[List[str]]` (batch of pretokenized examples)."
        )

        assert (
            text_pair is None
            or isinstance(text_pair, str)
            or (
                isinstance(text_pair, (list, tuple))
                and (
                    len(text_pair) == 0
                    or (
                        isinstance(text_pair[0], str)
                        or (
                            isinstance(text_pair[0], (list, tuple))
                            and (len(text_pair[0]) == 0 or isinstance(text_pair[0][0], str))
                        )
                    )
                )
            )
        ), ("text_pair input must of type `str` (single example), `List[str]` (batch or single pretokenized example) "
            "or `List[List[str]]` (batch of pretokenized examples).")

        is_batched = bool(
            (not is_split_into_words and isinstance(text, (list, tuple)))
            or (
                is_split_into_words and isinstance(text, (list, tuple)) and text and isinstance(text[0], (list, tuple))
            )
        )

        if is_batched:  # 判断输入的是一个batch还是单个序列
            batch_text_or_text_pairs = list(zip(text, text_pair)) if text_pair is not None else text
            return self.batch_encode_plus(
                batch_text_or_text_pairs=batch_text_or_text_pairs,
                add_special_tokens=add_special_tokens,
                padding=padding,
                truncation=truncation,
                max_length=max_length,
                stride=stride,
                is_split_into_words=is_split_into_words,
                return_tensors=return_tensors,
                return_overflowing_tokens=return_overflowing_tokens,
                return_special_tokens_mask=return_special_tokens_mask,
                return_offsets_mapping=return_offsets_mapping,
                **kwargs,
            )
        else:
            return self.encode_plus(
                text=text,
                text_pair=text_pair,
                add_special_tokens=add_special_tokens,
                padding=padding,
                truncation=truncation,
                max_length=max_length,
                stride=stride,
                is_split_into_words=is_split_into_words,
                return_tensors=return_tensors,
                return_overflowing_tokens=return_overflowing_tokens,
                return_special_tokens_mask=return_special_tokens_mask,
                return_offsets_mapping=return_offsets_mapping,
                **kwargs,
            )

    def encode_plus(
        self,
        text: Union[TextInput, PreTokenizedInput, EncodedInput],
        text_pair: Optional[Union[TextInput, PreTokenizedInput, EncodedInput]] = None,
        add_special_tokens: bool = True,
        padding: Union[bool, str, PaddingStrategy] = False,
        truncation: Union[bool, str, TruncationStrategy] = False,
        max_length: Optional[int] = None,
        stride: int = 0,
        is_split_into_words: bool = False,
        return_tensors: Optional[Union[str, TensorType]] = None,
        return_overflowing_tokens: bool = False,
        return_special_tokens_mask: bool = False,
        return_offsets_mapping: bool = False,
        **kwargs
    ) -> BatchEncoding:

        padding_strategy, truncation_strategy, max_length, kwargs = self._get_padding_truncation_strategies(
            padding=padding,
            truncation=truncation,
            max_length=max_length,
            **kwargs,
        )

        def get_input_ids(text):
            if isinstance(text, str):
                tokens = self.tokenize(text, **kwargs)
                return self.convert_tokens_to_ids(tokens)
            elif isinstance(text, (list, tuple)) and len(text) > 0 and isinstance(text[0], str):
                if is_split_into_words:
                    tokens = list(itertools.chain(*(self.tokenize(t, is_split_into_words=True, **kwargs) for t in text)))
                    return self.convert_tokens_to_ids(tokens)
                else:
                    return self.convert_tokens_to_ids(text)
            elif isinstance(text, (list, tuple)) and len(text) > 0 and isinstance(text[0], int):
                return text
            else:
                raise ValueError("Input is not valid. Should be a string, a list/tuple of strings or a list/tuple of integers.")

        first_offsets_mapping, second_offsets_mapping, offsets_mapping = None, None, None
        if return_offsets_mapping:
            first_offsets_mapping = self.rematch(text, self._tokenize(text))
            second_offsets_mapping = self.rematch(text_pair, self._tokenize(text_pair)) if text_pair is not None else None
        offsets_mapping = ((first_offsets_mapping, second_offsets_mapping)) if second_offsets_mapping else first_offsets_mapping

        first_ids = get_input_ids(text)
        second_ids = get_input_ids(text_pair) if text_pair is not None else None

        batch_outputs = self.prepare_for_model(
            first_ids=first_ids,
            second_ids=second_ids,
            add_special_tokens=add_special_tokens,
            padding_strategy=padding_strategy,
            truncation_strategy=truncation_strategy,
            offsets_mapping=offsets_mapping,
            max_length=max_length,
            stride=stride,
            return_overflowing_tokens=return_overflowing_tokens,
            return_special_tokens_mask=return_special_tokens_mask,
            return_offsets_mapping = return_offsets_mapping,
            return_tensors=return_tensors,
            prepend_batch_axis=False,
        )
        return batch_outputs


    def prepare_for_model(
        self,
        first_ids: List[int],
        second_ids: Optional[List[int]] = None,
        add_special_tokens: bool = True,
        padding_strategy: PaddingStrategy = PaddingStrategy.DO_NOT_PAD,
        truncation_strategy: TruncationStrategy = TruncationStrategy.DO_NOT_TRUNCATE,
        offsets_mapping=None,
        max_length: Optional[int] = None,
        stride: int = 0,
        return_tensors: Optional[Union[str, TensorType]] = None,
        return_overflowing_tokens: bool = False,
        return_special_tokens_mask: bool = False,
        return_offsets_mapping: bool = False,
        prepend_batch_axis: bool = False,
        **kwargs
    ) -> BatchEncoding:
        pair = bool(second_ids is not None)  # 输入的是句子还是句子对
        len_ids = len(first_ids)
        len_second_ids = len(second_ids) if pair else 0

        encoded_inputs = {}
        total_len = len_ids + len_second_ids + (self.num_special_tokens_to_add(pair=pair) if add_special_tokens else 0)

        overflowing_tokens = []
        if truncation_strategy != TruncationStrategy.DO_NOT_TRUNCATE and max_length and total_len > max_length:
            first_ids, second_ids, overflowing_tokens = self.truncate_sequences(
                first_ids,
                second_ids=second_ids,
                num_tokens_to_remove=total_len - max_length,
                truncation_strategy=truncation_strategy,
                stride=stride,
            )

        if return_overflowing_tokens:
            encoded_inputs["overflowing_tokens"] = overflowing_tokens
            encoded_inputs["num_truncated_tokens"] = total_len - max_length if total_len - max_length>=0 else 0

        if add_special_tokens:
            sequence = self.build_inputs_with_special_tokens(first_ids, second_ids)
            token_type_ids = self.create_token_type_ids_from_sequences(first_ids, second_ids)
            attention_mask = [1] * len(sequence)
        else:
            sequence = first_ids + second_ids if pair else first_ids
            token_type_ids = [0] * len(first_ids) + ([0] * len(second_ids) if pair else [])
            attention_mask = [1] * len(sequence)

        # Build output dictionary
        encoded_inputs["input_ids"] = sequence
        encoded_inputs["token_type_ids"] = token_type_ids
        encoded_inputs["attention_mask"] = attention_mask
        if return_special_tokens_mask:
            if add_special_tokens:
                encoded_inputs["special_tokens_mask"] = self.get_special_tokens_mask(first_ids, second_ids)
            else:
                encoded_inputs["special_tokens_mask"] = [0] * len(sequence)


        # Padding
        if padding_strategy != PaddingStrategy.DO_NOT_PAD:
            encoded_inputs = self.pad(encoded_inputs, max_length=max_length, padding_strategy=padding_strategy)

        if return_offsets_mapping:
            encoded_inputs['offsets_mapping'] = offsets_mapping


        batch_outputs = BatchEncoding(encoded_inputs, tensor_type=return_tensors, prepend_batch_axis=prepend_batch_axis)
        return batch_outputs

    def batch_encode_plus(
        self,
        batch_text_or_text_pairs: Union[
            List[TextInput],
            List[TextInputPair],
            List[PreTokenizedInput],
            List[PreTokenizedInputPair],
            List[EncodedInput],
            List[EncodedInputPair],
        ],
        add_special_tokens: bool = True,
        padding: Union[bool, str, PaddingStrategy] = False,
        truncation: Union[bool, str, TruncationStrategy] = False,
        max_length: Optional[int] = None,
        stride: int = 0,
        is_split_into_words: bool = False,
        return_tensors: Optional[Union[str, TensorType]] = None,
        return_overflowing_tokens: bool = False,
        return_special_tokens_mask: bool = False,
        return_offsets_mapping: bool = False,
        **kwargs
    ) -> BatchEncoding:

        padding_strategy, truncation_strategy, max_length, kwargs = self._get_padding_truncation_strategies(
            padding=padding,
            truncation=truncation,
            max_length=max_length,
            **kwargs)

        def get_input_ids(text):
            if isinstance(text, str):
                tokens = self.tokenize(text, **kwargs)
                return self.convert_tokens_to_ids(tokens)
            elif isinstance(text, (list, tuple)) and len(text) > 0 and isinstance(text[0], str):
                if is_split_into_words:
                    tokens = list(itertools.chain(*(self.tokenize(t, is_split_into_words=True, **kwargs) for t in text)))
                    return self.convert_tokens_to_ids(tokens)
                else:
                    return self.convert_tokens_to_ids(text)
            elif isinstance(text, (list, tuple)) and len(text) > 0 and isinstance(text[0], int):
                return text
            else:
                raise ValueError("Input is not valid. Should be a string, a list/tuple of strings or a list/tuple of integers.")

        offsets_mapping = [] if return_offsets_mapping else None
        input_ids = []
        for first_ids_or_second_ids in batch_text_or_text_pairs:
            if not isinstance(first_ids_or_second_ids, (list, tuple)):
                first_ids, second_ids = first_ids_or_second_ids, None
            elif is_split_into_words and not isinstance(first_ids_or_second_ids[0], (list, tuple)):
                first_ids, second_ids = first_ids_or_second_ids, None
            else:
                first_ids, second_ids = first_ids_or_second_ids

            if return_offsets_mapping:    # first_ids, second_ids此时为token
                first_ids_mapping = self.rematch(first_ids, self._tokenize(first_ids))
                second_ids_mapping = self.rematch(second_ids, self._tokenize(second_ids)) if second_ids is not None else None
                if second_ids is not None:
                    offsets_mapping.append((first_ids_mapping, second_ids_mapping))
                else:
                    offsets_mapping.append(first_ids_mapping)

            first_ids = get_input_ids(first_ids)
            second_ids = get_input_ids(second_ids) if second_ids is not None else None
            input_ids.append((first_ids, second_ids))   # token_id

        if padding_strategy == PaddingStrategy.BATCH_LONGEST:
            if add_special_tokens:
                batch_max_length = max([len(first_ids + second_ids)+3 if second_ids is not None else len(first_ids)+2
                                      for first_ids, second_ids in input_ids])
                max_length = max_length if batch_max_length > max_length else batch_max_length
            else:
                batch_max_length = max([len(first_ids + second_ids) if second_ids is not None else len(first_ids)
                                      for first_ids, second_ids in input_ids])
                max_length = max_length if batch_max_length > max_length else batch_max_length
        else:
            max_length = max_length

        batch_outputs = {}
        for first_ids, second_ids in input_ids:
            outputs = self.prepare_for_model(
                first_ids=first_ids,
                second_ids=second_ids,
                add_special_tokens=add_special_tokens,
                padding_strategy=padding_strategy,
                truncation_strategy=truncation_strategy,
                offsets_mapping=offsets_mapping,
                max_length=max_length,
                stride=stride,
                pad_to_multiple_of=None,
                return_offsets_mapping=return_offsets_mapping,
                return_overflowing_tokens=return_overflowing_tokens,
                return_special_tokens_mask=return_special_tokens_mask,
                return_tensors=None,  # We convert the whole batch to tensors at the end
                prepend_batch_axis=False)

            for key, value in outputs.items():
                if key not in batch_outputs:
                    batch_outputs[key] = []
                batch_outputs[key].append(value)

        batch_outputs = BatchEncoding(batch_outputs, tensor_type=return_tensors)
        return batch_outputs

    def pad(
        self,
        encoded_inputs: Union[
            BatchEncoding,
            List[BatchEncoding],
            Dict[str, EncodedInput],
            Dict[str, List[EncodedInput]],
            List[Dict[str, EncodedInput]],
        ],
        padding_strategy: Union[PaddingStrategy] = True,
        max_length: Optional[int] = None,
        return_tensors: Optional[Union[str, TensorType]] = None,
    ) -> BatchEncoding:

        if isinstance(encoded_inputs, (list, tuple)) and isinstance(encoded_inputs[0], (dict, BatchEncoding)):  # 如果是batch
            encoded_inputs = {key: [example[key] for example in encoded_inputs] for key in encoded_inputs[0].keys()}

        required_input = encoded_inputs["input_ids"]
        if required_input and not isinstance(required_input[0], (list, tuple)): # required_input[0]不是list,则代表输入的不是一个batch

            encoded_inputs = self._pad(
                encoded_inputs,
                max_length=max_length,
                padding_strategy=padding_strategy)
            return BatchEncoding(encoded_inputs, tensor_type=return_tensors)

        # batch_size = len(required_input)
        # assert all(
        #     len(v) == batch_size for v in encoded_inputs.values()
        # ), "Some items in the output dictionary have a different batch size than others."
        #
        # if padding_strategy == PaddingStrategy.LONGEST:
        #     max_length = max(len(inputs) for inputs in required_input)
        #     padding_strategy = PaddingStrategy.MAX_LENGTH
        #
        # batch_outputs = {}
        # for i in range(batch_size):
        #     inputs = dict((k, v[i]) for k, v in encoded_inputs.items())
        #     outputs = self._pad(inputs, max_length=max_length, padding_strategy=padding_strategy)
        #
        #     for key, value in outputs.items():
        #         if key not in batch_outputs:
        #             batch_outputs[key] = []
        #         batch_outputs[key].append(value)
        #
        # return BatchEncoding(batch_outputs, tensor_type=return_tensors)

    def create_token_type_ids_from_sequences(self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None) -> List[int]:
        if token_ids_1 is None:
            return len(token_ids_0) * [0]
        return [0] * len(token_ids_0) + [1] * len(token_ids_1)

    def build_inputs_with_special_tokens(self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None) -> List[int]:
        if token_ids_1 is None:
            return token_ids_0
        return token_ids_0 + token_ids_1

    def truncate_sequences(
        self,
        ids: List[int],
        second_ids: Optional[List[int]] = None,
        num_tokens_to_remove: int = 0,
        truncation_strategy: Union[str, TruncationStrategy] = "longest_first",
        stride: int = 0,
    ) -> Tuple[List[int], List[int], List[int]]:

        if num_tokens_to_remove <= 0:
            return ids, second_ids, []

        if not isinstance(truncation_strategy, TruncationStrategy):
            truncation_strategy = TruncationStrategy(truncation_strategy)

        overflowing_tokens = []
        if truncation_strategy == TruncationStrategy.LONGEST_FIRST:
            for _ in range(num_tokens_to_remove):
                if second_ids is None or len(ids) > len(second_ids):
                    if not overflowing_tokens:
                        window_len = min(len(ids), stride + 1)
                    else:
                        window_len = 1
                    overflowing_tokens.extend(ids[-window_len:])
                    ids = ids[:-1]
                else:
                    if not overflowing_tokens:
                        window_len = min(len(second_ids), stride + 1)
                    else:
                        window_len = 1
                    overflowing_tokens.extend(second_ids[-window_len:])
                    second_ids = second_ids[:-1]
        elif truncation_strategy == TruncationStrategy.ONLY_FIRST:
            if len(ids) > num_tokens_to_remove:
                window_len = min(len(ids), stride + num_tokens_to_remove)
                overflowing_tokens = ids[-window_len:]
                ids = ids[:-num_tokens_to_remove]

        elif truncation_strategy == TruncationStrategy.ONLY_SECOND and second_ids is not None:
            if len(second_ids) > num_tokens_to_remove:
                window_len = min(len(second_ids), stride + num_tokens_to_remove)
                overflowing_tokens = second_ids[-window_len:]
                second_ids = second_ids[:-num_tokens_to_remove]

        return (ids, second_ids, overflowing_tokens)

    def _pad(
        self,
        encoded_inputs: Union[Dict[str, EncodedInput], BatchEncoding],
        max_length: Optional[int] = None,
        padding_strategy: PaddingStrategy = PaddingStrategy.DO_NOT_PAD,
    ) -> dict:
        required_input = encoded_inputs["input_ids"]
        needs_to_be_padded = padding_strategy != PaddingStrategy.DO_NOT_PAD and len(required_input) < max_length

        if needs_to_be_padded:
            difference = max_length - len(required_input)
            encoded_inputs["attention_mask"] = encoded_inputs["attention_mask"] + [0] * difference
            encoded_inputs["token_type_ids"] = (encoded_inputs["token_type_ids"] + [self.pad_token_type_id] * difference)
            if "special_tokens_mask" in encoded_inputs:
                encoded_inputs["special_tokens_mask"] = encoded_inputs["special_tokens_mask"] + [1] * difference
            encoded_inputs["input_ids"] = required_input + [self.pad_token_id] * difference

        return encoded_inputs

    def batch_decode(
        self,
        sequences: Union[List[int], List[List[int]], "np.ndarray", "torch.Tensor"],
        skip_special_tokens: bool = False,
        clean_up_tokenization_spaces: bool = True,
        spaces_between_special_tokens: bool = True) -> List[str]:
        return [
            self.decode(
                seq,
                skip_special_tokens=skip_special_tokens,
                clean_up_tokenization_spaces=clean_up_tokenization_spaces,
                spaces_between_special_tokens=spaces_between_special_tokens,)
            for seq in sequences]

    def decode(
        self,
        token_ids: Union[int, List[int], "np.ndarray", "torch.Tensor"],
        skip_special_tokens: bool = False,
        clean_up_tokenization_spaces: bool = True,
        spaces_between_special_tokens: bool = True) -> str:

        token_ids = to_py_obj(token_ids)
        filtered_tokens = self.convert_ids_to_tokens(token_ids, skip_special_tokens=skip_special_tokens)

        sub_texts = []
        current_sub_text = []
        for token in filtered_tokens:
            if skip_special_tokens and token in self.all_special_ids:
                continue
            if token in self.added_tokens_encoder:
                if current_sub_text:
                    sub_texts.append(self.convert_tokens_to_string(current_sub_text))
                    current_sub_text = []
                sub_texts.append(token)
            else:
                current_sub_text.append(token)
        if current_sub_text:
            sub_texts.append(self.convert_tokens_to_string(current_sub_text))

        if spaces_between_special_tokens:
            text = " ".join(sub_texts)
        else:
            text = "".join(sub_texts)

        if clean_up_tokenization_spaces:
            clean_text = self.clean_up_tokenization(text)
            return clean_text
        else:
            return text

    def get_special_tokens_mask(self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None, already_has_special_tokens: bool = False) -> List[int]:
        """
        针对特殊token的mask，对应位置为1，否则为0
        """
        assert already_has_special_tokens and token_ids_1 is None, (
            "You cannot use ``already_has_special_tokens=False`` with this tokenizer. "
            "Please use a slow (full python) tokenizer to activate this argument."
            "Or set `return_special_tokens_mask=True` when calling the encoding method "
            "to get the special tokens mask in any tokenizer. "
        )
        all_special_ids = self.all_special_ids  # cache the property
        special_tokens_mask = [1 if token in all_special_ids else 0 for token in token_ids_0]

        return special_tokens_mask

    @staticmethod
    def clean_up_tokenization(out_string: str) -> str:
        """
        清理掉一些空格
        """
        out_string = (
            out_string.replace(" .", ".")
            .replace(" ?", "?")
            .replace(" !", "!")
            .replace(" ,", ",")
            .replace(" ' ", "'")
            .replace(" n't", "n't")
            .replace(" 'm", "'m")
            .replace(" 's", "'s")
            .replace(" 've", "'ve")
            .replace(" 're", "'re")
        )
        return out_string

    @staticmethod  # 该方法不强制要求传递参数
    def stem(token):  # 获取token的“词干”（如果是##开头，则自动去掉##）
            if token[:2] == '##':
                return token[2:]
            else:
                return token

    @staticmethod
    def _is_special(ch):  # 判断是不是有特殊含义的符号
            return bool(ch) and (ch[0] == '[') and (ch[-1] == ']')

    def rematch(self, text, tokens):
        """给出原始的text和tokenize后的tokens的映射关系
        """
        if self.do_lower_case:
            text = text.lower()

        normalized_text, char_mapping = '', []
        for i, ch in enumerate(text):
            if self.do_lower_case:
                ch = unicodedata.normalize('NFD', ch)
                ch = ''.join([c for c in ch if unicodedata.category(c) != 'Mn'])
            ch = ''.join([
                c for c in ch
                if not (ord(c) == 0 or ord(c) == 0xfffd or _is_control(c))
            ])
            normalized_text += ch
            char_mapping.extend([i] * len(ch))

        text, token_mapping, offset = normalized_text, [], 0
        for token in tokens:
            if self._is_special(token):
                token_mapping.append([])
            else:
                token = self.stem(token)
                start = text[offset:].index(token) + offset
                end = start + len(token)
                token_mapping.append(char_mapping[start:end])
                offset = end
        return token_mapping






import torch.nn as nn
from functools import partial
import warnings

import torch
from torch.utils.data import Dataset
from tqdm import tqdm
import time
from torch.cuda.amp import autocast as autocast
from fantasybert.core.trainer_utils import _build_args, _get_func_signature
from fantasybert.core.utils import _move_dict_value_to_device
from fantasybert.core.data_utils import BatchIter, TrainerState
from fantasybert.core.trainer_utils import _model_contains_inner_module, _data_parallel_wrapper
from fantasybert.core.logger_utils import logger
from fantasybert.core.loss_main import prepare_losser, LossBase
from fantasybert.core.metrics import _prepare_metrics, MetricBase
warnings.filterwarnings("ignore")
from typing import Union, Any


class Tester(object):
    def __init__(self,
                 model: nn.Module,
                 args: Union[TrainerState, Any],
                 data: Union[Dataset, BatchIter],
                 loss: Union[LossBase, str] = 'cross_entropy',
                 metrics: Union[MetricBase, str] = None,
                 collate_fn = None,
                 shuffle: bool = False,
                 use_tqdm: bool = False,
                 **kwargs):

        super(Tester, self).__init__()

        if not isinstance(model, nn.Module):
            raise TypeError(f"The type of model must be `torch.nn.Module`, got `{type(model)}`.")

        self.args = args
        self.data = data
        self.device = self.args.device
        self._model = model.to(self.device)
        self.batch_size = self.args.eval_batch_size
        self.use_tqdm = use_tqdm
        self.logger = logger
        self.pin_memory = kwargs.get('pin_memory', True)

        self.metrics = _prepare_metrics(metrics)
        self.eval_loss = loss
        if loss != 'none':
            self.losser = prepare_losser(loss)

        # data
        if isinstance(self.data, Dataset):
            self.data_iterator = BatchIter(dataset=self.data, batch_size=self.batch_size, sampler=self.args.sampler,
                                           num_workers=self.args.num_workers, drop_last=self.args.drop_last,
                                           collate_fn=collate_fn, shuffle=shuffle,
                                           use_block_shuffle=self.args.eval_use_block_shuffle, **kwargs)
        elif isinstance(self.data, BatchIter):
            self.data_iterator = self.data
            self.data = self.data.dataset
        else:
            raise TypeError("train_data type {} not support".format(type(self.data)))

        # check predict
        if (hasattr(self._model, 'predict') and callable(self._model.evaluate)) or \
                (_model_contains_inner_module(self._model) and hasattr(self._model.module, 'predict') and
                 callable(self._model.module.predict)):
            if isinstance(self._model, nn.DataParallel):
                self._predict_func_wrapper = partial(_data_parallel_wrapper('predict',
                                                                            self._model.device_ids,
                                                                            self._model.output_device),
                                                     network=self._model.module)
                self._predict_func = self._model.module.predict  # 用于匹配参数
            elif isinstance(self._model, nn.parallel.DistributedDataParallel):
                self._predict_func = self._model.module.predict
                self._predict_func_wrapper = self._model.module.predict  # 用于调用
            else:
                self._predict_func = self._model.evaluate
                self._predict_func_wrapper = self._model.evaluate
        else:
            if _model_contains_inner_module(model):
                self._predict_func_wrapper = self._model.forward
                self._predict_func = self._model.module.forward
            else:
                self._predict_func = self._model.forward
                self._predict_func_wrapper = self._model.forward


    def _compute_loss(self, predict, truth, batch_data):
        return self.losser(predict, truth, batch_data, self.args)


    def test(self):
        r"""开始进行验证，并返回验证结果。

        :return Dict[Dict]: dict的二层嵌套结构，dict的第一层是metric的名称; 第二层是这个metric的指标。一个AccuracyMetric的例子为{'AccuracyMetric': {'acc': 1.0}}。
        """
        # turn on the testing mode; clean up the history
        self._model.eval()
        data_iterator = self.data_iterator
        eval_results = {}
        avg_loss = 0

        with torch.no_grad():
            if not self.use_tqdm:
                from fantasybert.core.trainer_utils import _pseudo_tqdm as inner_tqdm
            else:
                inner_tqdm = tqdm
            with inner_tqdm(total=len(data_iterator), leave=False, dynamic_ncols=True) as pbar:
                pbar.set_description_str(desc="Test")

                start_time = time.time()
                for batch in self.data_iterator:
                    batch_x, batch_y = self._get_batch_xy(batch)

                    _move_dict_value_to_device(batch_x, batch_y, device=self.device)
                    _move_dict_value_to_device(batch, device=self.device)

                    pred_dict = self._data_forward(self._predict_func, batch)

                    if self.eval_loss != 'none':
                        loss = self._compute_loss(pred_dict, batch_y, batch_x).mean()
                        avg_loss += loss.item()

                    if not isinstance(pred_dict, dict):
                        raise TypeError(f"The return value of {_get_func_signature(self._predict_func)} "
                                        f"must be `dict`, got {type(pred_dict)}.")
                    for metric in self.metrics:
                        metric(pred_dict, batch_y)

                    if self.use_tqdm:
                        pbar.update()

                avg_loss = float(avg_loss) / len(self.data_iterator)
                for metric in self.metrics:
                    eval_result = metric.get_metric()
                    if not isinstance(eval_result, dict):
                        raise TypeError(f"The return value of {_get_func_signature(metric.get_metric)} must be "
                                        f"`dict`, got {type(eval_result)}")
                    metric_name = metric.get_metric_name()
                    eval_results[metric_name] = eval_result
                if self.eval_loss != 'none':
                    eval_results['loss'] = {'loss': round(avg_loss,7)}
                pbar.close()
                end_time = time.time()
                test_str = f'Evaluate data in {round(end_time - start_time, 2)} seconds!'
                self.logger.info(test_str)

        return eval_results

    def _get_batch_xy(self, batch):
        batch_x, batch_y = {}, None
        for name, item in batch.items():
            if name == 'label':
                batch_y = {}
                batch_y['target'] = batch['label']
            else:
                batch_x[name] = item
        return batch_x, batch_y

    def _data_forward(self, func, x):
        x = _build_args(func, **x)
        if self.args.fp16:
            with autocast():
                outputs = self._predict_func_wrapper(**x)
        else:
            outputs = self._predict_func_wrapper(**x)
        return outputs

    def _format_eval_results(self, results):
        _str = ''
        for metric_name, metric_result in results.items():
            _str += metric_name + ': '
            _str += ", ".join([str(key) + "=" + str(value) for key, value in metric_result.items()])
            _str += '\n'
        return _str[:-1]



#! -*- coding: utf-8 -*-
import os
import torch
from torch.utils.data import Dataset
from itertools import chain
from torch.utils.data.dataloader import _SingleProcessDataLoaderIter, _MultiProcessingDataLoaderIter
from torch.utils.data import DataLoader
from torch.utils.data import Sampler, SequentialSampler
import random, abc
import json
import dataclasses
from typing import Any, Optional, List, Tuple, Union, Dict
from fantasybert.tokenization.tokenizers import BertTokenizer

from dataclasses import dataclass, field, asdict
def set_seed(args):
    SEED = args.SEED
    torch.manual_seed(SEED)
    torch.cuda.manual_seed_all(SEED)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = True


@dataclass
class TrainerState:
    # must passed in parameters
    learning_rate: float = field(default_factory=float)
    n_epochs: int = field(default_factory=int)
    max_len: int = field(default_factory=int)
    batch_size: int = field(default_factory=int)
    model_path: str = field(default_factory=str)

    save_path: Optional[str] = None
    data_dir: Optional[str] = None
    tokenizer: BertTokenizer = None


    train_batch_size: int = None
    eval_batch_size: int = None
    SEED: int = 42
    metric_key: str = "accuracy"
    patience: int = 5

    global_step: int = 0
    max_steps: int = 0
    test_use_tqdm = False

    save_only_param: bool = True

    # block_shuffle_args: Dict[bool, int, Any] = field(default_factory=set_block_shuffle_args)
    use_block_shuffle: bool = False
    eval_use_block_shuffle: bool = False
    batch_in_shuffle: bool = True
    sort_bs_num: Optional[int] = None   # Optional[X] 等价于 Union[X, None]
    sort_key = None

    num_labels: int = None
    add_polynominal: bool = False

    save_step_model: bool = False

    best_metric: Optional[float] = None

    fp16: bool = False
    use_adv: Union[bool, str] = False
    use_ema: bool = False
    use_lookahead: bool = False

    gradient_clip: bool = True
    clip_value = 1.0
    clip_type: str = 'norm'

    gradient_accumulation_steps: int = 1

    warmup_ratios: float = 0.1
    weight_decay: float = 1e-3

    load_checkpoint: bool = False

    # dataloader
    num_workers = 0
    sampler = None
    drop_last: bool = False
    train_shuffle: bool = True
    eval_shuffle: bool = False
    pin_memory = None


    device: Union[torch.device, str] = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    def __getstate__(self):
        return self.__dict__

    def __post_init__(self):
        if isinstance(self.device, str):
            self.device = torch.device(self.device)

        if self.sort_key is None:
            self.sort_key = lambda x: len(self.tokenizer.tokenize(x[0]))  # x[0] is text in Dataset

        if self.save_path is not None:
            os.makedirs(self.save_path, exist_ok=True)

        if self.batch_size is not None and self.train_batch_size is None and self.eval_batch_size is None:
            self.train_batch_size = self.eval_batch_size = self.batch_size

    def save_to_json(self, json_path: str):
        json_string = json.dumps(dataclasses.asdict(self), indent=2, sort_keys=True) + "\n"
        with open(json_path, "w", encoding="utf-8") as f:
            f.write(json_string)

    @classmethod
    def load_from_json(cls, json_path: str):
        with open(json_path, "r", encoding="utf-8") as f:
            text = f.read()
        return cls(**json.loads(text))

    @property
    def to_dict(self):
        return asdict(self)


class DataSetGetter(Dataset):
    def __init__(self, file_path=None, datas=None):
        if isinstance(file_path, (str, list)):
            self.datas, self.total_labels = self.load_data(file_path)
        elif isinstance(datas, list):
            self.datas = datas
        else:
            raise ValueError('The input args shall be str format file_path / list format datas')
        print(f'Num samples of {file_path} is {len(self.datas)}')

    def __len__(self):
        return len(self.datas)

    def __getitem__(self, idx):
        return self.datas[idx]

    @staticmethod
    def load_data(file_path):
        return file_path

    @property
    def data(self):
        return self.datas

    @property
    def num_labels(self):
        return len(list(set(self.total_labels))) if self.total_labels is not None and self.total_labels != [] else None

    @property
    def label2id(self):
        return {l:i for i,l in enumerate(sorted(list(set(self.total_labels))))}\
            if self.total_labels is not None and self.total_labels != [] else None

    @property
    def id2label(self):
        return {i:l for i,l in enumerate(sorted(list(set(self.total_labels))))}\
            if self.total_labels is not None and self.total_labels != [] else None

    @property
    def labels_distribution(self):
        if self.total_labels is not None and self.total_labels != []:
            labels_dic = {}
            for label in self.total_labels:
                labels_dic[label] = labels_dic.get(label, 0) + 1
            total_num = sum(list(labels_dic.values()))
            label_distribution = dict((x, round((y/total_num)*100, 3)) for x, y in labels_dic.items())
            sorted_label_distribution = dict(sorted(label_distribution.items(), key=lambda x: -float(x[1])))
            final_label_distribution = {k: str(v) + '%' for k, v in sorted_label_distribution.items()}
            return final_label_distribution
        else:
            return None


class BatchIter(DataLoader):
    def __init__(self,
                 dataset: Union[DataSetGetter, Dataset],
                 sort_bs_num=None,
                 sort_key=None,
                 use_block_shuffle: bool = False,
                 batch_in_shuffle: bool = False,
                 batch_size=1, sampler=None,
                 num_workers=0, pin_memory=False, drop_last=False,
                 timeout=0, worker_init_fn=None, collate_fn=None,
                 batch_sampler=None, shuffle=False,
                 **kwargs,
                 ):
        batch_sampler = batch_sampler
        if batch_sampler is not None:
            kwargs['batch_size'] = 1
            kwargs['sampler'] = None
            kwargs['drop_last'] = False

        super().__init__(dataset=dataset, batch_size=batch_size, sampler=sampler,
            collate_fn=collate_fn, num_workers=num_workers,
            pin_memory=pin_memory, drop_last=drop_last,
            timeout=timeout, worker_init_fn=worker_init_fn,
            batch_sampler=batch_sampler, shuffle=shuffle)

        assert len(dataset) > 0, 'dataset cannot be None'
        assert isinstance(dataset.datas, list), "the data attribute of DatasetGetter object must be a list"

        self.use_block_shuffle = use_block_shuffle
        self.sort_bs_num = sort_bs_num
        self.sort_key = sort_key
        self.batch_in_is_shuffle = batch_in_shuffle

    def __iter__(self):
        if self.use_block_shuffle is False:
            if self.num_workers == 0:
                return _SingleProcessDataLoaderIter(self)
            else:
                return _MultiProcessingDataLoaderIter(self)

        if self.use_block_shuffle is True:
            # self.dataset is the attribute in torch DataLoader
            self.dataset.datas = self.block_shuffle(self.dataset.datas, self.batch_size, self.sort_bs_num,
                                                       self.sort_key, self.batch_in_is_shuffle)
            if self.num_workers == 0:
                return _SingleProcessDataLoaderIter(self)
            else:
                return _MultiProcessingDataLoaderIter(self)

    @staticmethod
    def block_shuffle(data, batch_size, sort_bs_num, sort_key, batch_in_shuffle):
        random.shuffle(data)
        # 将数据按照batch_size大小进行切分
        tail_data = [] if len(data) % batch_size == 0 else data[-(len(data) % batch_size):]
        data = data[:len(data) - len(tail_data)]
        assert len(data) % batch_size == 0
        # 获取真实排序范围
        sort_bs_num = len(data) // batch_size if sort_bs_num is None else sort_bs_num
        # 按照排序范围进行数据划分
        data = [data[i:i + sort_bs_num * batch_size] for i in range(0, len(data), sort_bs_num * batch_size)]
        # 在排序范围，根据排序函数进行降序排列
        data = [sorted(i, key=sort_key, reverse=True) for i in data]
        # 将数据根据batch_size获取batch_data
        data = list(chain(*data))
        data = [data[i:i + batch_size] for i in range(0, len(data), batch_size)]
        # 判断是否需要对batch_data序列进行打乱
        if batch_in_shuffle:
            random.shuffle(data)
        # 将tail_data填补回去
        data = list(chain(*data)) + tail_data
        return data








import os
from torch.utils.data import DataLoader
import warnings
from pkg_resources import parse_version
import torch
import torch.nn as nn
from torch.utils.data import Dataset
from tqdm import tqdm
import time, datetime
from torch.optim.lr_scheduler import LambdaLR
from torch.optim import Optimizer
from torch.cuda.amp import autocast as autocast, GradScaler
from fantasybert.models.modeling_utils import build_bert_optim_parameters
from fantasybert.core.tester import Tester
from fantasybert.core.trainer_utils import _build_args, _get_func_signature, _model_contains_inner_module, _save_model
from fantasybert.core.utils import _move_dict_value_to_device
from fantasybert.core.data_utils import BatchIter, TrainerState
from fantasybert.core.callback import CallbackManager, Callback, CallbackException
from fantasybert.core.logger_utils import logger
from fantasybert.core.loss_main import prepare_losser, LossBase
from fantasybert.core.metrics import _prepare_metrics, MetricBase
from fantasybert.core.optimization import _prepare_optimizers, get_scheduler, Lookahead
warnings.filterwarnings("ignore")
from typing import Union, Optional, Any, List


class Trainer(object):
    def __init__(self,
                 model: nn.Module,
                 args: Union[TrainerState, Any],
                 train_data: Union[Dataset, BatchIter, DataLoader],
                 tester: Optional[Tester] = None,
                 loss: Union[LossBase, str] = 'cross_entropy',
                 eval_loss: Union[LossBase, str] = 'as_train',
                 optimizer: Union[Optimizer, str] = 'AdamW',
                 scheduler: Union[LambdaLR, str] = None,
                 dev_data: Union[Dataset, BatchIter, DataLoader] = None,
                 test_data: Union[Dataset, BatchIter, DataLoader] = None,
                 callbacks: Union[List[Callback], List[str], Callback, str] = 'default',
                 stay_default_callback: bool = False, # set stay_default_callback = True only when we pass new callbacks, therefore callbacks!='default'
                 metrics: Union[List[MetricBase], List[str], MetricBase, str] = None,
                 collate_fn=None,
                 train_collate_fn=None,
                 eval_collate_fn=None,
                 train_shuffle: bool = True,
                 eval_shuffle: bool = False,
                 use_tqdm: bool = True,
                 validate_every: int = -1,
                 print_every: int = -1,
                 show_train_result=True,
                 **kwargs,
                 ):
        super(Trainer, self).__init__()

        self.args = args
        if collate_fn is not None and train_collate_fn is None and eval_collate_fn is None:
            self.train_collate_fn = collate_fn
            self.eval_collate_fn = collate_fn
        else:
            self.train_collate_fn = train_collate_fn
            self.eval_collate_fn = eval_collate_fn
        self.train_shuffle = train_shuffle
        self.eval_shuffle = eval_shuffle

        self.batch_size = self.args.train_batch_size
        self.fp16 = self.args.fp16
        self.n_epochs = self.args.n_epochs
        self.use_block_shuffle = self.args.use_block_shuffle

        self.validate_every = validate_every
        self.show_train_result = show_train_result

        self.train_data = train_data
        self.dev_data = dev_data
        self.test_data = test_data
        self.use_tqdm = use_tqdm
        self.stay_default_callback = stay_default_callback

        self.best_metric_indicator = None
        self.best_dev_epoch = None
        self.best_dev_step = None
        self.best_dev_perf = None

        self.device = self.args.device
        self.model = model
        self.model.to(self.device)

        self.pin_memory = kwargs.get('pin_memory', False if parse_version(torch.__version__)==parse_version('1.9') else True)
        self.print_every = print_every if print_every != -1 else self.validate_every
        self.set_grad_to_none = kwargs.get('set_grad_to_none', True)

        self.scaler = GradScaler()

        # model
        if not isinstance(model, nn.Module):
            raise TypeError(f"The type of model must be torch.nn.Module, got {type(model)}.")

        # data
        self.worker_init_fn = kwargs.get('worker_init_fn')
        self.batch_sampler = kwargs.get('batch_sampler')
        if isinstance(train_data, Dataset):
            self.data_iterator = BatchIter(dataset=train_data, batch_size=self.batch_size, sampler=self.args.sampler,
                                             num_workers=self.args.num_workers, drop_last=self.args.drop_last,
                                            collate_fn=self.train_collate_fn, shuffle=self.train_shuffle,
                                                use_block_shuffle=self.use_block_shuffle,**kwargs)
        elif isinstance(train_data, DataLoader):
            self.data_iterator = train_data
        elif isinstance(train_data, BatchIter):
            self.data_iterator = train_data
        else:
            raise TypeError("train_data type {} not support".format(type(train_data)))

        if _model_contains_inner_module(self.model):
            self._forward_func = self.model.module.forward
        else:
            self._forward_func = self.model.forward

        # gradient_accumulation_steps
        self.gradient_accumulation_steps = self.args.gradient_accumulation_steps
        assert self.gradient_accumulation_steps >= 1, "gradient_accumulation_steps must be no less than 1."

        self.n_steps = (len(self.data_iterator) // self.gradient_accumulation_steps) * self.n_epochs
        # TODO：到底用不用除？

        # optimizer
        self.optimizer_clss = _prepare_optimizers(optimizer)
        optimizer_grouped_parameters = build_bert_optim_parameters(self.model, self.args.weight_decay)
        self.optimizer = self.optimizer_clss(optimizer_grouped_parameters, lr=self.args.learning_rate)
        if self.args.use_lookahead:
            self.optimizer = Lookahead(optimizer=self.optimizer, **self.args.optim_args)

        # scheduler
        self.warmup_steps = int(self.n_steps * self.args.warmup_ratios)
        if isinstance(scheduler, str):
            self.scheduler = get_scheduler(name=scheduler,
                                           optimizer=self.optimizer,
                                           num_warmup_steps=self.warmup_steps,
                                           num_training_steps=self.n_steps)
        elif isinstance(scheduler, LambdaLR):
            self.scheduler = scheduler
        else:
            self.scheduler = None

        # loss
        self.losser = prepare_losser(loss)
        if eval_loss == 'as_train':
            self.eval_loss = loss
        else:
            self.eval_loss = eval_loss

        # metrics
        self.metrics = _prepare_metrics(metrics)

        # metric_key
        self.metric_key = self.args.metric_key
        if self.metric_key is not None:
            self.metric_key = self.metric_key[1:] if self.metric_key[0] == "+" or self.metric_key[0] == "-" else self.metric_key
        else:
            self.metric_key = None


        # tester
        if self.dev_data is not None:
            if tester is None:
                self.tester = Tester(model=self.model,
                                 args=self.args,
                                 loss=self.eval_loss,
                                 data=self.dev_data,
                                 metrics=self.metrics,
                                 collate_fn=self.eval_collate_fn,
                                 shuffle=self.eval_shuffle,
                                 use_tqdm=self.args.test_use_tqdm)
            else:
                self.tester = tester

        # save_path
        self.save_path = self.args.save_path
        if not (self.args.save_path is None or isinstance(self.args.save_path, str)):
            raise ValueError("save_path can only be None or `str`.")

        # callback
        if isinstance(callbacks, Callback):
            callbacks = [callbacks]
        self.callback_manager = CallbackManager(env={"trainer": self},
                                                callbacks=callbacks)


    def _compute_loss(self, predict, truth, batch_data):
        return self.losser(predict, truth, batch_data, self.args)


    def train(self, load_best_model=True, on_exception='auto', **kwargs):
        results = {}

        if self.n_epochs <= 0:
            logger.info(f"training epoch is {self.n_epochs}, nothing was done.")
            results['seconds'] = 0.
            return results
        try:
            self.model.train()
            self._load_best_model = load_best_model
            self.start_time = str(datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S'))
            start_time = time.time()
            logger.info("training epochs started " + self.start_time)
            self.step = 0
            self.epoch = 1

            try:
                self.callback_manager.on_train_begin()
                self._train()
                self.callback_manager.on_train_end()

            except BaseException as e:
                self.callback_manager.on_exception(e)
                if on_exception == 'auto':
                    if not isinstance(e, (CallbackException, KeyboardInterrupt)):
                        raise e
                elif on_exception == 'raise':
                    raise e


            if self.dev_data is not None and self.best_dev_perf is not None and load_best_model:
                model_name = "best_" + "_".join([self.model.__class__.__name__, self.metric_key])
                load_succeed = self._load_model(self.model, model_name)
                if load_succeed:
                    logger.info("Reloaded the best model.")
                else:
                    logger.info("Fail to reload best model.")

            if self.dev_data is None and self.save_path is not None:
                model_name = "_".join([self.model.__class__.__name__])
                _save_model(self.model, model_name, self.save_path, self.device, self.args.save_only_param)

        finally:
            if self.dev_data is not None and self.best_dev_perf is not None:
                print(self.best_dev_perf)
                logger.info(
                    "\nIn Epoch:{}/Step:{}, got best dev performance:".format(self.best_dev_epoch, self.best_dev_step))
                logger.info(self.tester._format_eval_results(self.best_dev_perf))
                results['best_eval'] = self.best_dev_perf
                results['best_epoch'] = self.best_dev_epoch
                results['best_step'] = self.best_dev_step

        results['seconds'] = round(time.time() - start_time, 2)

        return results

    def _train(self):
        if not self.use_tqdm:
            from .trainer_utils import _pseudo_tqdm as inner_tqdm
        else:
            inner_tqdm = tqdm
        start = time.time()
        with inner_tqdm(total=self.n_steps, postfix='loss:{0:<6.5f}', leave=False, dynamic_ncols=True,
                        initial=self.step) as pbar:
            self.pbar = pbar
            avg_loss = 0
            self.batch_per_epoch = len(self.data_iterator)
            for epoch in range(self.epoch, self.n_epochs + 1):
                self.epoch = epoch
                pbar.set_description_str(desc="Epoch {}/{}".format(epoch, self.n_epochs))
                # early stopping
                self.callback_manager.on_epoch_begin()
                for batch in self.data_iterator:
                    self.step += 1
                    batch_x, batch_y = self._get_batch_xy(batch)
                    if batch_y is not None:
                        _move_dict_value_to_device(batch_x, batch_y, device=self.device)
                    else:
                        _move_dict_value_to_device(batch_x, device=self.device)
                    _move_dict_value_to_device(batch, device=self.device)

                    # negative sampling; replace unknown; re-weight batch_y
                    self.callback_manager.on_batch_begin(batch_x, batch_y)
                    prediction = self._data_forward(self.model, batch)

                    # edit prediction
                    self.callback_manager.on_loss_begin(batch_y, prediction)
                    loss = self._compute_loss(prediction, batch_y, batch_x).mean()
                    loss = loss / self.gradient_accumulation_steps
                    avg_loss += loss.item()

                    # Is loss NaN or inf? requires_grad = False
                    self.callback_manager.on_backward_begin(loss)
                    self._grad_backward(loss)

                    self.callback_manager.on_backward_end(batch_x, batch_y)

                    self._update()
                    self.callback_manager.on_step_end()
                    self._gradient_zero_grad()

                    if self.step % self.print_every == 0:
                        avg_loss = float(avg_loss) / self.print_every
                        if self.use_tqdm:
                            print_output = "loss:{:<6.5f}".format(avg_loss)
                            pbar.update(self.print_every)
                        else:
                            end = time.time()
                            diff = round(end - start)
                            print_output = "[epoch: {:>3} step: {:>4}] train loss: {:>4.6} time: {}".format(
                                epoch, self.step, avg_loss, diff)
                        pbar.set_postfix_str(print_output)
                        avg_loss = 0
                    self.callback_manager.on_batch_end()

                    if (self.validate_every > 0 and self.step % self.validate_every == 0) \
                            and self.dev_data is not None:
                        eval_res = self._do_validation(epoch=epoch, step=self.step)
                        eval_str = "Evaluation on dev at Epoch {}/{}. Step:{}/{}: ".format(epoch, self.n_epochs, self.step,
                                                                                    self.n_steps)
                        # pbar.write(eval_str + '\n')
                        logger.info(eval_str)
                        logger.info(self.tester._format_eval_results(eval_res)+'\n')
                # ================= mini-batch end ==================== #
                if self.validate_every<0 and self.dev_data is not None:  # 在epoch结束之后的evaluate
                    eval_res = self._do_validation(epoch=epoch, step=self.step)
                    eval_str = "Evaluation on dev at Epoch {}/{}. Step:{}/{}: ".format(epoch, self.n_epochs, self.step,
                                                                                       self.n_steps)
                    # pbar.write(eval_str + '\n')
                    logger.info(eval_str)
                    logger.info(self.tester._format_eval_results(eval_res) + '\n')
                # lr decay; early stopping
                self.callback_manager.on_epoch_end()
            # =============== epochs end =================== #
            if self.dev_data is not None and (self.validate_every>0 and self.n_steps%self.validate_every!=0):
                eval_res = self._do_validation(epoch=epoch, step=self.step)
                eval_str = "Evaluation on dev at Epoch {}/{}. Step:{}/{}: ".format(epoch, self.n_epochs, self.step,
                                                                                   self.n_steps)
                # pbar.write(eval_str + '\n')
                logger.info(eval_str)
                logger.info(self.tester._format_eval_results(eval_res) + '\n')
            pbar.close()
            self.pbar = None
        # ============ tqdm end ============== #

    def _data_forward(self, network, x):
        x = _build_args(self._forward_func, **x)
        if self.fp16:
            # 前向过程(model + loss)开启 autocast
            with autocast():
                outputs = network(**x)
        else:
            outputs = network(**x)
        if not isinstance(outputs, dict):
            raise TypeError(
                f"The return value of {_get_func_signature(self._forward_func)} should be dict, got {type(outputs)}.")
        return outputs

    def _grad_backward(self, loss):
        if self.fp16:
            # Scales loss，这是因为半精度的数值范围有限，因此需要用它放大
            self.scaler.scale(loss).backward()
        else:
            loss.backward()  #  gradient_accumulation_steps？？

    def _gradient_zero_grad(self):
        if self.step % self.gradient_accumulation_steps == 0:
            self._clear_grad(self.optimizer, self.set_grad_to_none)

    def _clear_grad(self, optimizer, set_to_none=True):
        param_groups = optimizer.param_groups
        for group in param_groups:
            for p in group['params']:
                if p.grad is not None:
                    if set_to_none:
                        p.grad = None
                    else:
                        if p.grad.grad_fn is not None:
                            p.grad.detach_()
                        else:
                            p.grad.requires_grad_(False)
                        p.grad.zero_()

    def _update(self):
        if self.fp16:
            self.scaler.unscale_(self.optimizer)
            # scaler.step() unscale之前放大后的梯度，但是scale太多可能出现inf或NaN
            # 故其会判断是否出现了inf/NaN
            # 如果梯度的值不是 infs 或者 NaNs, 那么调用optimizer.step()来更新权重,
            # 如果检测到出现了inf或者NaN，就跳过这次梯度更新，同时动态调整scaler的大小
            self.scaler.step(self.optimizer)
            # 查看是否要更新scaler
            self.scaler.update()
        else:
            self.optimizer.step()

    def _do_validation(self, epoch, step):
        self.callback_manager.on_valid_begin()
        res = self.tester.test()

        is_better_eval = False
        if self._better_eval_result(res):
            if self.save_path is not None:
                _save_model(self.model, "best_" + "_".join([self.model.__class__.__name__, self.metric_key]),
                            self.save_path, self.device, self.args.save_only_param)
            elif self._load_best_model:
                self._best_model_states = {name: param.cpu().clone() for name, param in self.model.state_dict().items()}
            self.best_dev_perf = res
            self.best_dev_epoch = epoch
            self.best_dev_step = step
            is_better_eval = True
        # get validation results; adjust optimizer
        self.callback_manager.on_valid_end(res, self.metric_key, self.optimizer, is_better_eval)
        return res

    def test(self, model_name):
        load_succeed = self._load_model(self.model, model_name)
        if load_succeed:
            logger.info("Successfully loaded the best model.")
        else:
            logger.info("Failed to load best model.")
        tester = Tester(model=self.model,
               args=self.args,
               data=self.test_data,
               metrics=self.metrics,
               collate_fn=self.eval_collate_fn,
               shuffle=self.eval_shuffle,
               use_tqdm=True)
        res = tester.test()
        for k,v in res.items():
            print(v)


    def _better_eval_result(self, metrics):
        r"""Check if the current epoch yields better validation results.

        :return bool value: True means current results on dev set is the best.
        """
        indicator, indicator_val = _check_eval_results(metrics, self.metric_key, self.metrics)
        if self.metric_key is None:
            self.metric_key = indicator
        is_better = True
        if self.best_metric_indicator is None:
            # first-time validation
            self.best_metric_indicator = indicator_val
        else:
            if indicator != 'loss':
                if indicator_val > self.best_metric_indicator:
                    self.best_metric_indicator = indicator_val
                else:
                    is_better = False
            else:
                if indicator_val < self.best_metric_indicator:
                    self.best_metric_indicator = indicator_val
                else:
                    is_better = False
        return is_better


    def _load_model(self, model, model_name, only_param=False):
        # 返回bool值指示是否成功reload模型
        if self.save_path is not None:
            model_path = os.path.join(self.save_path, model_name)
            if only_param:
                states = torch.load(model_path)
            else:
                states = torch.load(model_path)
                #states = torch.load(model_path).state_dict()
            if _model_contains_inner_module(model):
                model.module.load_state_dict(states)
            else:
                model.load_state_dict(states)
        elif hasattr(self, "_best_model_states"):
            model.load_state_dict(self._best_model_states)
        else:
            return False
        return True


    def _get_batch_xy(self, batch):
        batch_x, batch_y = {}, None
        for name, item in batch.items():
            if name == 'label':
                batch_y = {}
                batch_y['target'] = batch['label']
            else:
                batch_x[name] = item
        return batch_x, batch_y


def _check_eval_results(metrics, metric_key, metric_list):
    # metrics: tester返回的结果
    # metric_key: 一个用来做筛选的指标，来自Trainer的初始化
    # metric_list: 多个用来做评价的指标，来自Trainer的初始化
    if isinstance(metrics, tuple):
        loss, metrics = metrics

    indicator_val, indicator = None, None
    if isinstance(metrics, dict):

        if metric_key is None:
            metric_dict = list(metrics.values())[0]  # 取第一个metric
            indicator_val, indicator = list(metric_dict.values())[0], list(metric_dict.keys())[0]
        else:
            metric_dict = list(metrics.values())
            # [{'accuracy': 0.073444}, {'f1': 0.073448, 'precision': 0.073451, 'recall': 0.073444}, {'loss': 3.644089}]
            if not any(metric_key in _.keys() for _ in metric_dict):
                raise RuntimeError(f"metric key {metric_key} not found in {metric_dict}")

            for item in metric_dict:
                if metric_key in item.keys():
                    indicator_val = item[metric_key]
                    indicator = metric_key
    else:
        raise RuntimeError("Invalid metrics type. Expect {}, got {}".format((tuple, dict), type(metrics)))

    return indicator, indicator_val


#! -*- coding: utf-8 -*-

import os, sys, six, re, json
from collections import OrderedDict, UserDict
from dataclasses import fields
import inspect
import torch
import time
import functools
import torch.nn.functional as F
import json
import os
import torch.nn as nn
from enum import Enum
import random
import numpy as np
from collections import defaultdict
import torch
from dataclasses import dataclass, field, asdict
import inspect
from fantasybert.core.logger_utils import logger
_open_ = open
is_py2 = six.PY2

if not is_py2:
    basestring = str


@dataclass
class Const:
    INPUT = 'words'
    CHAR_INPUT = 'chars'
    INPUT_LEN = 'seq_len'
    OUTPUT = 'pred'
    TARGET = 'target'
    LOSS = 'loss'
    RAW_WORD = 'raw_words'
    RAW_CHAR = 'raw_chars'

    @staticmethod
    def INPUTS(i):
        r"""得到第 i 个 ``INPUT`` 的命名"""
        i = int(i) + 1
        return Const.INPUT + str(i)

    @staticmethod
    def CHAR_INPUTS(i):
        r"""得到第 i 个 ``CHAR_INPUT`` 的命名"""
        i = int(i) + 1
        return Const.CHAR_INPUT + str(i)

    @staticmethod
    def RAW_WORDS(i):
        r"""得到第 i 个 ``RAW_WORDS`` 的命名"""
        i = int(i) + 1
        return Const.RAW_WORD + str(i)

    @staticmethod
    def RAW_CHARS(i):
        r"""得到第 i 个 ``RAW_CHARS`` 的命名"""
        i = int(i) + 1
        return Const.RAW_CHAR + str(i)

    @staticmethod
    def INPUT_LENS(i):
        r"""得到第 i 个 ``INPUT_LEN`` 的命名"""
        i = int(i) + 1
        return Const.INPUT_LEN + str(i)

    @staticmethod
    def OUTPUTS(i):
        r"""得到第 i 个 ``OUTPUT`` 的命名"""
        i = int(i) + 1
        return Const.OUTPUT + str(i)

    @staticmethod
    def TARGETS(i):
        r"""得到第 i 个 ``TARGET`` 的命名"""
        i = int(i) + 1
        return Const.TARGET + str(i)

    @staticmethod
    def LOSSES(i):
        r"""得到第 i 个 ``LOSS`` 的命名"""
        i = int(i) + 1
        return Const.LOSS + str(i)


def seq_len_to_mask(seq_len, max_len=None):
    r"""

    将一个表示sequence length的一维数组转换为二维的mask，不包含的位置为0。
    转变 1-d seq_len到2-d mask.

    .. code-block::

        >>> seq_len = torch.arange(2, 16)
        >>> mask = seq_len_to_mask(seq_len)
        >>> print(mask.size())
        torch.Size([14, 15])
        >>> seq_len = np.arange(2, 16)
        >>> mask = seq_len_to_mask(seq_len)
        >>> print(mask.shape)
        (14, 15)
        >>> seq_len = torch.arange(2, 16)
        >>> mask = seq_len_to_mask(seq_len, max_len=100)
        >>>print(mask.size())
        torch.Size([14, 100])

    :param np.ndarray,torch.LongTensor seq_len: shape将是(B,)
    :param int max_len: 将长度pad到这个长度。默认(None)使用的是seq_len中最长的长度。但在nn.DataParallel的场景下可能不同卡的seq_len会有
        区别，所以需要传入一个max_len使得mask的长度是pad到该长度。
    :return: np.ndarray, torch.Tensor 。shape将是(B, max_length)， 元素类似为bool或torch.uint8
    """
    if isinstance(seq_len, np.ndarray):
        assert len(np.shape(seq_len)) == 1, f"seq_len can only have one dimension, got {len(np.shape(seq_len))}."
        max_len = int(max_len) if max_len else int(seq_len.max())
        broad_cast_seq_len = np.tile(np.arange(max_len), (len(seq_len), 1))
        mask = broad_cast_seq_len < seq_len.reshape(-1, 1)

    elif isinstance(seq_len, torch.Tensor):
        assert seq_len.dim() == 1, f"seq_len can only have one dimension, got {seq_len.dim() == 1}."
        batch_size = seq_len.size(0)
        max_len = int(max_len) if max_len else seq_len.max().long()
        broad_cast_seq_len = torch.arange(max_len).expand(batch_size, -1).to(seq_len)
        mask = broad_cast_seq_len.lt(seq_len.unsqueeze(1))
    else:
        raise TypeError("Only support 1-d numpy.ndarray or 1-d torch.Tensor.")

    return mask



# calculate running time for the function
def log_execution_time(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start = time.perf_counter()
        res = func(*args, **kwargs)
        end = time.perf_counter()
        print('{} took {} ms'.format(func.__name__, (end - start) * 1000))
        return res

    return wrapper


def batch_sequence_padding(inputs, length=None, value=0, seq_dims=1, mode='post'):
    if length is None:
        length = np.max([np.shape(x)[:seq_dims] for x in inputs], axis=0)
    elif not hasattr(length, '__getitem__'):
        length = [length]

    slices = [np.s_[:length[i]] for i in range(seq_dims)]
    slices = tuple(slices) if len(slices) > 1 else slices[0]
    pad_width = [(0, 0) for _ in np.shape(inputs[0])]

    outputs = []
    for x in inputs:
        x = x[slices]
        for i in range(seq_dims):
            if mode == 'post':
                pad_width[i] = (0, length[i] - np.shape(x)[i])
            elif mode == 'pre':
                pad_width[i] = (length[i] - np.shape(x)[i], 0)
            else:
                raise ValueError('"mode" argument must be "post" or "pre".')
        x = np.pad(x, pad_width, 'constant', constant_values=value)
        outputs.append(x)
    return np.array(outputs)


def insert_arguments(**arguments):
    """装饰器，为类方法增加参数
    （主要用于类的__init__方法）
    """
    def actual_decorator(func):
        def new_func(self, *args, **kwargs):
            for k, v in arguments.items():
                if k in kwargs:
                    v = kwargs.pop(k)
                setattr(self, k, v)
            return func(self, *args, **kwargs)

        return new_func

    return actual_decorator


def delete_arguments(*arguments):
    """装饰器，为类方法删除参数
    （主要用于类的__init__方法）
    """
    def actual_decorator(func):
        def new_func(self, *args, **kwargs):
            for k in arguments:
                if k in kwargs:
                    raise TypeError(
                        '%s got an unexpected keyword argument \'%s\'' %
                        (self.__class__.__name__, k)
                    )
            return func(self, *args, **kwargs)

        return new_func

    return actual_decorator


def longest_common_substring(source, target):
    """最长公共子串（source和target的最长公共切片区间）
    返回：子串长度, 所在区间（四元组）
    注意：最长公共子串可能不止一个，所返回的区间只代表其中一个。
    """
    c, l, span = defaultdict(int), 0, (0, 0, 0, 0)
    for i, si in enumerate(source, 1):
        for j, tj in enumerate(target, 1):
            if si == tj:
                c[i, j] = c[i - 1, j - 1] + 1
                if c[i, j] > l:
                    l = c[i, j]
                    span = (i - l, i, j - l, j)
    return l, span


def longest_common_subsequence(source, target):
    """最长公共子序列（source和target的最长非连续子序列）
    返回：子序列长度, 映射关系（映射对组成的list）
    注意：最长公共子序列可能不止一个，所返回的映射只代表其中一个。
    """
    c = defaultdict(int)
    for i, si in enumerate(source, 1):
        for j, tj in enumerate(target, 1):
            if si == tj:
                c[i, j] = c[i - 1, j - 1] + 1
            elif c[i, j - 1] > c[i - 1, j]:
                c[i, j] = c[i, j - 1]
            else:
                c[i, j] = c[i - 1, j]
    l, mapping = c[len(source), len(target)], []
    i, j = len(source) - 1, len(target) - 1
    while len(mapping) < l:
        if source[i] == target[j]:
            mapping.append((i, j))
            i, j = i - 1, j - 1
        elif c[i + 1, j] > c[i, j + 1]:
            j = j - 1
        else:
            i = i - 1
    return l, mapping[::-1]


def l2_normalize(vecs):
    """标准化
    """
    norms = (vecs ** 2).sum(axis=1, keepdims=True) ** 0.5
    return vecs / np.clip(norms, 1e-8, np.inf)


def compute_corrcoef(x, y):
    """Spearman相关系数
    """
    import scipy.stats
    return scipy.stats.spearmanr(x, y).correlation


class DataAugment(object):
    def __init__(self, ):
        self.PUNCTUATIONS = ['.', ',', ' ', '。', '，', '、', '？']

    def insert_punctuation_marks(self, sentence, punc_ratio=0.3):
        import jieba
        words = [word for word in jieba.cut(sentence)]  # jieba分词，往词中间随机加标点
        new_line = []
        q = random.randint(1, int(punc_ratio * len(words) + 1))
        qs = random.sample(range(0, len(words)), q)

        for j, word in enumerate(words):
            if j in qs:
                new_line.append(self.PUNCTUATIONS[random.randint(0, len(self.PUNCTUATIONS) - 1)])
                new_line.append(word)
            else:
                new_line.append(word)
        new_line = ' '.join(new_line)
        return new_line

    def aeda_zh(self, input_text, punc_ratio=0.3):
        if isinstance(input_text, list) and len(input_text) > 1:
            total = []
            for text in input_text:
                total.append(self.insert_punctuation_marks(text, punc_ratio))
            return total

        elif isinstance(input_text, tuple) and len(input_text) > 1:
            total = []
            for text in input_text:
                total.append(self.insert_punctuation_marks(text, punc_ratio))
            return total

        elif isinstance(input_text, str):
            return self.insert_punctuation_marks(input_text, punc_ratio)

        else:
            raise AttributeError('Can only get input type of list, tuple and str')

    @classmethod
    def token_level_cutoff(self, input_ids, attention_mask, cutoff_rate=0.15, cutoff_type='token'):
        bsz, seq_len = input_ids.shape
        _input_ids = []
        cutoff_attention_mask = []

        for bsz_id in range(bsz):
            sample_mask = attention_mask[bsz_id]
            _input_id = None
            # num_tokens = sample_mask.sum().int().item()
            num_tokens = sample_mask.sum().item()  # 当前句子中的token数
            cur_input_id = input_ids[bsz_id]  # 当前的input_ids

            if cutoff_type == 'span':
                sample_len = max(int(num_tokens * (1 - cutoff_rate)),
                                 1)  # if true_len is 32, cutoff_rate is 0.15 then sample_len is 27
                start_id = np.random.randint(1, high=num_tokens - sample_len + 1)  # 从1开始选择span，避免删除CLS, 同时也要避免超过句子长度
                cutoff_mask = [1] * seq_len
                for idx in range(start_id, start_id + sample_len):
                    cutoff_mask[idx] = 0  # 这些位置是0，bool之后就变成了False，而masked_fill是选择True的位置替换为value的

                cutoff_mask[0] = 0  # 避免CLS被替换
                cutoff_mask[num_tokens - 1] = 0  # 避免SEP被替换

            elif cutoff_type == 'token':
                cutoff_token_num = max(int(num_tokens * cutoff_rate), 1)
                cutoff_mask = [0] * seq_len
                idxs = random.sample([idx for idx in range(1, num_tokens - 1)], cutoff_token_num)
                # 不重复的选取cutoff_token_num个token的idx，同时避免选取CLS，SEP
                for idx in idxs:
                    cutoff_mask[idx] = 1  # 1对应True，masked_fill会把对应True的位置替换为0

            else:
                raise TypeError('Please type the right cutoff type!!!')

            cutoff_mask = torch.ByteTensor(cutoff_mask).bool().cuda()
            _input_id = cur_input_id.masked_fill(cutoff_mask, value=0).cuda()
            sample_mask = sample_mask.masked_fill(cutoff_mask, value=0).cuda()

            _input_id = _input_id.view(1, -1)  # 增加一个维度
            sample_mask = sample_mask.view(1, -1)

            _input_ids.append(_input_id)
            cutoff_attention_mask.append(sample_mask)

        _input_ids = torch.cat(_input_ids, dim=0)
        cutoff_attention_mask = torch.cat(cutoff_attention_mask, dim=0)
        return _input_ids, cutoff_attention_mask

    @classmethod
    def vec_level_cutoff(self, embeddings, cutoff_rate=0.05):
        bs, hidden_size = embeddings.shape
        cutoff_embed_num = max(int(hidden_size * cutoff_rate), 1)
        cut_embeddings = []
        for bs_id in range(bs):
            cutoff_mask = [1] * hidden_size
            embed = embeddings[bs_id]
            idxs = random.sample([idx for idx in range(hidden_size)], cutoff_embed_num)
            for idx in idxs:
                cutoff_mask[idx] = 0  # 1原来的向量值不变，0则相乘为0

            cutoff_mask = torch.tensor(cutoff_mask).cuda()
            embed = embed * cutoff_mask
            embed = embed.view(1, -1)
            cut_embeddings.append(embed)

        cut_embeddings = torch.cat(cut_embeddings, dim=0)
        return cut_embeddings

    # TODO: eng有点问题，应该空格分开,另添加标点符号，token cutoff等
    def word_repetition_eng(self, input_text, dup_rate):
        def repetition_for_one(text, dup_rate):
            actual_len = len(text)
            dup_len = random.randint(a=0, b=max(
                2, int(dup_rate * actual_len)))
            dup_word_index = random.sample(
                list(range(1, actual_len)), k=dup_len)

            dup_text = ''
            for index, word in enumerate(text):
                dup_text += word
                if index in dup_word_index:
                    dup_text += word
            return dup_text

        if isinstance(input_text, list) and len(input_text) > 1:
            total = []
            for text in input_text:
                total.append(repetition_for_one(text, dup_rate))
            return total

        elif isinstance(input_text, tuple) and len(input_text) > 1:
            total = []
            for text in input_text:
                total.append(repetition_for_one(text, dup_rate))
            return total

        elif isinstance(input_text, str):
            return repetition_for_one(input_text, dup_rate)

        else:
            raise AttributeError('Can only get input type of list, tuple and str')

    @classmethod
    def word_repetition_chinese(self, input_text, dup_rate):
        ''' span duplicated for chinese
        '''
        import jieba
        def repetition_for_one(text, dup_rate):
            ori_text = text
            cut_text = jieba.cut(text, cut_all=False)
            text = list(cut_text)
            actual_len = len(text)
            if actual_len <= 4:
                return ori_text

            dup_len = random.randint(a=0, b=max(2, int(dup_rate * actual_len)))
            dup_word_index = random.sample(list(range(1, actual_len)), k=dup_len)

            dup_text = ''
            for index, word in enumerate(text):
                dup_text += word
                if index in dup_word_index:
                    dup_text += word
            return dup_text

        if isinstance(input_text, list) and len(input_text) > 1:
            total = []
            for text in input_text:
                total.append(repetition_for_one(text, dup_rate))
            return total

        elif isinstance(input_text, tuple) and len(input_text) > 1:
            total = []
            for text in input_text:
                total.append(repetition_for_one(text, dup_rate))
            return total

        elif isinstance(input_text, str):
            return repetition_for_one(input_text, dup_rate)

        else:
            raise AttributeError('Can only get input type of list, tuple and str')


def get_kw(cls, kwargs):
    '''保留排除cls的入参后的kwargs
    '''
    kwargs_new = {}
    for k in kwargs:
        if k not in set(inspect.getargspec(cls)[0]):
            kwargs_new[k] = kwargs[k]
    return kwargs_new



def kl_distance(p_logit, q_logit):
    p = F.softmax(p_logit, dim=-1)
    _kl = torch.sum(p * (F.log_softmax(p_logit, dim=-1)
                         - F.log_softmax(q_logit, dim=-1)), -1)
    return torch.mean(_kl)



class _pseudo_tqdm:
    r"""
    当无法引入tqdm，或者Trainer中设置use_tqdm为false的时候，用该方法打印数据
    """

    def __init__(self, **kwargs):
        self.logger = logger

    def write(self, info):
        self.logger.info(info)

    def set_postfix_str(self, info):
        self.logger.info(info)

    def __getattr__(self, item):
        def pass_func(*args, **kwargs):
            pass

        return pass_func

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        del self


def _move_dict_value_to_device(*args, device: torch.device, non_blocking=False):
    r"""

    move data to model's device, element in *args should be dict. This is a inplace change.
    :param device: torch.device
    :param non_blocking: bool, 是否异步将数据转移到cpu, 需要tensor使用pin_memory()
    :param args:
    :return:
    """
    if not torch.cuda.is_available() or device is None:
        return

    if not isinstance(device, torch.device):
        raise TypeError(f"device must be `torch.device`, got `{type(device)}`")

    for arg in args:
        if isinstance(arg, dict):
            for key, value in arg.items():
                if isinstance(value, torch.Tensor):
                    arg[key] = value.to(device, non_blocking=non_blocking)
        else:
            raise TypeError(f"Only support `dict` type right now got type {type(arg)}")


class ExplicitEnum(Enum):
    @classmethod  # classmethod修饰符对应的函数不需要实例化类，不需要self参数，但第一个参数需要是表示自身类的 cls 参数
    def _missing_(cls, value):
        raise ValueError(
            f"{value} is not a valid {cls.__name__}, please select one of {list(cls._value2member_map_.keys())}"
        )



def _is_function_or_method(func):
    r"""

    :param func:
    :return:
    """
    if not inspect.ismethod(func) and not inspect.isfunction(func):
        return False
    return True


def _check_function_or_method(func):
    if not _is_function_or_method(func):
        raise TypeError(f"{type(func)} is not a method or function.")



from typing import Optional, Union
from collections import Iterable
from numbers import Real
import warnings

import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import Optional

# based on:
# https://github.com/kevinzakka/pytorch-goodies/blob/master/losses.py
# adapted from:
# https://kornia.readthedocs.io/en/v0.1.2/_modules/torchgeometry/losses/dice.html


class DiceLoss(nn.Module):
    r"""Criterion that computes Sørensen-Dice Coefficient loss.

    According to [1], we compute the Sørensen-Dice Coefficient as follows:

    .. math::

        \text{Dice}(x, class) = \frac{2 |X| \cap |Y|}{|X| + |Y|}

    where:
       - :math:`X` expects to be the scores of each class.
       - :math:`Y` expects to be the one-hot tensor with the class labels.

    the loss, is finally computed as:

    .. math::

        \text{loss}(x, class) = 1 - \text{Dice}(x, class)

    [1] https://en.wikipedia.org/wiki/S%C3%B8rensen%E2%80%93Dice_coefficient

    Shape:
        - Input: :math:`(N, C)` where C = number of classes.
        - Target: :math:`(N,)` where each value is
          :math:`0 ≤ targets[i] ≤ C−1`.

    Examples:
        >>> N = 5  # num_classes
        >>> loss = DiceLoss()
        >>> input = torch.randn(2, N, requires_grad=True)
        >>> target = torch.empty(2, dtype=torch.long).random_(N)
        >>> output = loss(input, target)
        >>> output.backward()
    """

    def __init__(
        self,
        gamma: int = 0,
        scale: float = 1.0,
        reduction: Optional[str] = "mean",
        ignore_index: int = -100,
        eps: float = 1e-6,
        smooth: float = 0,
    ) -> None:
        super(DiceLoss, self).__init__()
        self.gamma: int = gamma
        self.scale: float = scale
        self.reduction: Optional[str] = reduction
        self.ignore_index: int = ignore_index
        self.eps: float = eps
        self.smooth: float = smooth

    def forward(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:

        if len(input.shape) == 2:
            if input.shape[0] != target.shape[0]:
                raise ValueError(
                    "number of elements in input and target shapes must be the same. Got: {}".format(
                        input.shape, input.shape
                    )
                )
        else:
            raise ValueError(
                "Invalid input shape, we expect or NxC. Got: {}".format(input.shape)
            )
        if not input.device == target.device:
            raise ValueError(
                "input and target must be in the same device. Got: {}".format(
                    input.device, target.device
                )
            )
        # compute softmax over the classes axis
        input_soft = F.softmax(input, dim=1)

        input_soft = self.scale * ((1 - input_soft) ** self.gamma) * input_soft

        # filter labels
        target = target.type(torch.long)
        input_mask = target != self.ignore_index

        target = target[input_mask]
        input_soft = input_soft[input_mask]

        # create the labels one hot tensor
        target_one_hot = (
            F.one_hot(target, num_classes=input_soft.shape[-1])
            .to(input.device)
            .type(input_soft.dtype)
        )

        # compute the actual dice score
        intersection = torch.sum(input_soft * target_one_hot, dim=-1)
        cardinality = torch.sum(input_soft + target_one_hot, dim=-1)

        dice_score = (2.0 * intersection + self.smooth) / (
            cardinality + self.eps + self.smooth
        )
        dice_loss = 1.0 - dice_score

        if self.reduction is None or self.reduction == "none":
            return dice_loss
        elif self.reduction == "mean":
            return torch.mean(dice_loss)
        elif self.reduction == "sum":
            return torch.sum(dice_loss)
        else:
            raise NotImplementedError(
                "Invalid reduction mode: {}".format(self.reduction)
            )


# based on:
# https://github.com/zhezh/focalloss/blob/master/focalloss.py
# adapted from:
# https://kornia.readthedocs.io/en/v0.1.2/_modules/torchgeometry/losses/focal.html


class FocalLoss(nn.Module):
    r"""Criterion that computes Focal loss.
    According to [1], the Focal loss is computed as follows:
    .. math::
        \text{FL}(p_t) = -\alpha_t (1 - p_t)^{\gamma} \, \text{log}(p_t)
    where:
       - :math:`p_t` is the model's estimated probability for each class.
    Arguments:
        alpha (float): Weighting factor :math:`\alpha \in [0, 1]` for one-vs-others mode (weight of negative class)
                        or :math:`\alpha_i \in \R`
                        vector of weights for each class (analogous to weight argument for CrossEntropyLoss)
        gamma (float): Focusing parameter :math:`\gamma >= 0`. When 0 is equal to CrossEntropyLoss
        reduction (Optional[str]): Specifies the reduction to apply to the
         output: ‘none’ | ‘mean’ | ‘sum’.
         ‘none’: no reduction will be applied,
         ‘mean’: the sum of the output will be divided by the number of elements
                in the output, uses geometric mean if alpha set to list of weights
         ‘sum’: the output will be summed. Default: ‘none’.
        ignore_index (Optional[int]): specifies indexes that are ignored during loss calculation
         (identical to PyTorch's CrossEntropyLoss 'ignore_index' parameter). Default: -100

    Shape:
        - Input: :math:`(N, C)` where C = number of classes.
        - Target: :math:`(N)` where each value is
          :math:`0 ≤ targets[i] ≤ C−1`.
    Examples:
        >>> C = 5  # num_classes
        >>> N = 1 # num_examples
        >>> loss = FocalLoss(alpha=0.5, gamma=2.0, reduction='mean')
        >>> input = torch.randn(N, C, requires_grad=True)
        >>> target = torch.empty(N, dtype=torch.long).random_(N)
        >>> output = loss(input, target)
        >>> output.backward()
    References:
        [1] https://arxiv.org/abs/1708.02002
    """

    def __init__(
        self,
        alpha: Optional[Union[float, Iterable]] = None,
        gamma: Real = 2.0,
        reduction: str = "mean",
        ignore_index: int = -100,
    ) -> None:
        super(FocalLoss, self).__init__()
        if (
            alpha is not None
            and not isinstance(alpha, float)
            and not isinstance(alpha, Iterable)
        ):
            raise ValueError(
                f"alpha value should be None, float value or list of real values. Got: {type(alpha)}"
            )
        self.alpha: Optional[Union[float, torch.Tensor]] = (
            alpha
            if alpha is None or isinstance(alpha, float)
            else torch.FloatTensor(alpha)
        )
        if isinstance(alpha, float) and not 0.0 <= alpha <= 1.0:
            warnings.warn("[Focal Loss] alpha value is to high must be between [0, 1]")

        self.gamma: Real = gamma
        self.reduction: str = reduction
        self.ignore_index: int = ignore_index

    def forward(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        if not torch.is_tensor(input):
            raise TypeError(
                "Input type is not a torch.Tensor. Got {}".format(type(input))
            )
        if input.shape[0] != target.shape[0]:
            raise ValueError(
                f"First dimension of inputs and targets should be same shape. "
                f"Got: {input.shape} and {target.shape}"
            )
        if len(input.shape) != 2 or len(target.shape) != 1:
            raise ValueError(
                f"input tensors should be of shape (N, C) and (N,). "
                f"Got: {input.shape} and {target.shape}"
            )
        if input.device != target.device:
            raise ValueError(
                "input and target must be in the same device. Got: {}".format(
                    input.device, target.device
                )
            )

        # filter labels
        target = target.type(torch.long)
        input_mask = target != self.ignore_index
        target = target[input_mask]
        input = input[input_mask]
        # compute softmax over the classes axis
        pt = F.softmax(input, dim=1)
        logpt = F.log_softmax(input, dim=1)

        # compute focal loss
        pt = pt.gather(1, target.unsqueeze(-1)).squeeze()
        logpt = logpt.gather(1, target.unsqueeze(-1)).squeeze()
        focal_loss = -1 * (1 - pt) ** self.gamma * logpt

        weights = torch.ones_like(
            focal_loss, dtype=focal_loss.dtype, device=focal_loss.device
        )
        if self.alpha is not None:
            if isinstance(self.alpha, float):
                alpha = torch.tensor(self.alpha, device=input.device)
                weights = torch.where(target > 0, 1 - alpha, alpha)
            elif torch.is_tensor(self.alpha):
                alpha = self.alpha.to(input.device)
                weights = alpha.gather(0, target)

        tmp_loss = focal_loss * weights
        if self.reduction == "none":
            loss = tmp_loss
        elif self.reduction == "mean":
            loss = (
                tmp_loss.sum() / weights.sum()
                if torch.is_tensor(self.alpha)
                else torch.mean(tmp_loss)
            )
        elif self.reduction == "sum":
            loss = tmp_loss.sum()
        else:
            raise NotImplementedError(
                "Invalid reduction mode: {}".format(self.reduction)
            )
        return loss



class RDropLoss(nn.Module):
    '''R-Drop的Loss实现，官方项目：https://github.com/dropreg/R-Drop
    '''
    def __init__(self, ce_loss, alpha):
        super().__init__()
        self.alpha = alpha
        self.loss_sup = ce_loss
        self.loss_rdrop = nn.KLDivLoss(reduction='none')

    def forward(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        pred1, pred2 = input
        loss_sup = self.loss_sup(pred1, target) + self.loss_sup(pred2, target)

        loss_rdrop1 = self.loss_rdrop(F.log_softmax(pred1, dim=-1), F.softmax(pred2, dim=-1))
        loss_rdrop2 = self.loss_rdrop(F.log_softmax(pred2, dim=-1), F.softmax(pred1, dim=-1))
        return loss_sup + (torch.mean(loss_rdrop1 + loss_rdrop2) / 4) * self.alpha


import collections
import os
import unicodedata
from typing import List, Optional
from fantasybert.tokenization.tokenization_base import TokenizerBase
from fantasybert.tokenization.tokenization_utils import _is_whitespace, _is_control, _is_punctuation

def load_vocab(vocab_file):
    vocab = collections.OrderedDict()
    with open(vocab_file, "r", encoding="utf-8") as reader:
        tokens = reader.readlines()
    for index, token in enumerate(tokens):
        token = token.rstrip("\n")
        vocab[token] = index
    return vocab


def whitespace_tokenize(text):
    """Runs basic whitespace cleaning and splitting on a piece of text."""
    text = text.strip()
    if not text:
        return []
    tokens = text.split()
    return tokens


class BertTokenizer(TokenizerBase):
    def __init__(
        self,
        vocab_file,
        do_lower_case=True,
        do_basic_tokenize=True,
        never_split=None,
        unk_token="[UNK]",
        sep_token="[SEP]",
        pad_token="[PAD]",
        cls_token="[CLS]",
        mask_token="[MASK]",
        tokenize_chinese_chars=True,
        strip_accents=None,
        **kwargs
    ):
        super().__init__(
            do_lower_case=do_lower_case,
            do_basic_tokenize=do_basic_tokenize,
            never_split=never_split,
            unk_token=unk_token,
            sep_token=sep_token,
            pad_token=pad_token,
            cls_token=cls_token,
            mask_token=mask_token,
            tokenize_chinese_chars=tokenize_chinese_chars,
            strip_accents=strip_accents,
            **kwargs,
        )

        if not os.path.isfile(vocab_file):
            raise ValueError(
                f"Can't find a vocabulary file at path '{vocab_file}'. To load the vocabulary from a Google pretrained "
                "model use `tokenizer = BertTokenizer.from_pretrained(PRETRAINED_MODEL_NAME)`"
            )
        self.vocab = load_vocab(vocab_file)
        self.ids_to_tokens = collections.OrderedDict([(ids, tok) for tok, ids in self.vocab.items()])
        self.do_basic_tokenize = do_basic_tokenize
        if do_basic_tokenize:
            self.basic_tokenizer = BasicTokenizer(
                do_lower_case=do_lower_case,
                never_split=never_split,
                tokenize_chinese_chars=tokenize_chinese_chars,
                strip_accents=strip_accents,
            )
        self.wordpiece_tokenizer = WordpieceTokenizer(vocab=self.vocab, unk_token=self.unk_token)


    @property
    def do_lower_case(self):
        return self.basic_tokenizer.do_lower_case

    @property
    def vocab_size(self):
        return len(self.vocab)

    def get_vocab(self):
        return dict(self.vocab, **self.added_tokens_encoder)

    def _tokenize(self, text):
        split_tokens = []
        if self.do_basic_tokenize:
            for token in self.basic_tokenizer.tokenize(text, never_split=self.all_special_tokens):

                # If the token is part of the never_split set
                if token in self.basic_tokenizer.never_split:
                    split_tokens.append(token)
                else:
                    split_tokens += self.wordpiece_tokenizer.tokenize(token)
        else:
            split_tokens = self.wordpiece_tokenizer.tokenize(text)
        return split_tokens

    def _convert_token_to_id(self, token):
        return self.vocab.get(token, self.vocab.get(self.unk_token))

    def _convert_id_to_token(self, index):
        return self.ids_to_tokens.get(index, self.unk_token)

    def convert_tokens_to_string(self, tokens):
        out_string = " ".join(tokens).replace(" ##", "").strip()
        return out_string

    def build_inputs_with_special_tokens(
        self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None
    ) -> List[int]:
        if token_ids_1 is None:
            return [self.cls_token_id] + token_ids_0 + [self.sep_token_id]
        cls = [self.cls_token_id]
        sep = [self.sep_token_id]
        return cls + token_ids_0 + sep + token_ids_1 + sep

    def get_special_tokens_mask(
        self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None, already_has_special_tokens: bool = False
    ) -> List[int]:
        if already_has_special_tokens:
            return super().get_special_tokens_mask(
                token_ids_0=token_ids_0, token_ids_1=token_ids_1, already_has_special_tokens=True
            )

        if token_ids_1 is not None:
            return [1] + ([0] * len(token_ids_0)) + [1] + ([0] * len(token_ids_1)) + [1]
        return [1] + ([0] * len(token_ids_0)) + [1]

    def create_token_type_ids_from_sequences(
        self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None
    ) -> List[int]:
        sep = [self.sep_token_id]
        cls = [self.cls_token_id]
        if token_ids_1 is None:
            return len(cls + token_ids_0 + sep) * [0]
        return len(cls + token_ids_0 + sep) * [0] + len(token_ids_1 + sep) * [1]


class BasicTokenizer(object):
    def __init__(self, do_lower_case=True, never_split=None, tokenize_chinese_chars=True, strip_accents=None):
        if never_split is None:
            never_split = []
        self.do_lower_case = do_lower_case
        self.never_split = set(never_split)
        self.tokenize_chinese_chars = tokenize_chinese_chars
        self.strip_accents = strip_accents

    def tokenize(self, text, never_split=None):
        never_split = self.never_split.union(set(never_split)) if never_split else self.never_split
        text = self._clean_text(text)

        if self.tokenize_chinese_chars:
            text = self._tokenize_chinese_chars(text)
        orig_tokens = whitespace_tokenize(text)
        split_tokens = []
        for token in orig_tokens:
            if token not in never_split:
                if self.do_lower_case:
                    token = token.lower()
                    if self.strip_accents is not False:
                        token = self._run_strip_accents(token)
                elif self.strip_accents:
                    token = self._run_strip_accents(token)
            split_tokens.extend(self._run_split_on_punc(token, never_split))

        output_tokens = whitespace_tokenize(" ".join(split_tokens))
        return output_tokens

    def _run_strip_accents(self, text):
        """Strips accents from a piece of text."""
        text = unicodedata.normalize("NFD", text)
        output = []
        for char in text:
            cat = unicodedata.category(char)
            if cat == "Mn":
                continue
            output.append(char)
        return "".join(output)

    def _run_split_on_punc(self, text, never_split=None):
        """Splits punctuation on a piece of text."""
        if never_split is not None and text in never_split:
            return [text]
        chars = list(text)
        i = 0
        start_new_word = True
        output = []
        while i < len(chars):
            char = chars[i]
            if _is_punctuation(char):
                output.append([char])
                start_new_word = True
            else:
                if start_new_word:
                    output.append([])
                start_new_word = False
                output[-1].append(char)
            i += 1

        return ["".join(x) for x in output]

    def _tokenize_chinese_chars(self, text):
        """Adds whitespace around any CJK character."""
        output = []
        for char in text:
            cp = ord(char)
            if self._is_chinese_char(cp):
                output.append(" ")
                output.append(char)
                output.append(" ")
            else:
                output.append(char)
        return "".join(output)

    def _is_chinese_char(self, cp):
        if (
            (cp >= 0x4E00 and cp <= 0x9FFF)
            or (cp >= 0x3400 and cp <= 0x4DBF)  #
            or (cp >= 0x20000 and cp <= 0x2A6DF)  #
            or (cp >= 0x2A700 and cp <= 0x2B73F)  #
            or (cp >= 0x2B740 and cp <= 0x2B81F)  #
            or (cp >= 0x2B820 and cp <= 0x2CEAF)  #
            or (cp >= 0xF900 and cp <= 0xFAFF)
            or (cp >= 0x2F800 and cp <= 0x2FA1F)  #
        ):  #
            return True

        return False

    def _clean_text(self, text):
        """Performs invalid character removal and whitespace cleanup on text."""
        output = []
        for char in text:
            cp = ord(char)
            if cp == 0 or cp == 0xFFFD or _is_control(char):
                continue
            if _is_whitespace(char):
                output.append(" ")
            else:
                output.append(char)
        return "".join(output)


class WordpieceTokenizer(object):
    def __init__(self, vocab, unk_token, max_input_chars_per_word=100):
        self.vocab = vocab
        self.unk_token = unk_token
        self.max_input_chars_per_word = max_input_chars_per_word

    def tokenize(self, text):
        output_tokens = []
        for token in whitespace_tokenize(text):
            chars = list(token)
            if len(chars) > self.max_input_chars_per_word:
                output_tokens.append(self.unk_token)
                continue

            is_bad = False
            start = 0
            sub_tokens = []
            while start < len(chars):
                end = len(chars)
                cur_substr = None
                while start < end:
                    substr = "".join(chars[start:end])
                    if start > 0:
                        substr = "##" + substr
                    if substr in self.vocab:
                        cur_substr = substr
                        break
                    end -= 1
                if cur_substr is None:
                    is_bad = True
                    break
                sub_tokens.append(cur_substr)
                start = end

            if is_bad:
                output_tokens.append(self.unk_token)
            else:
                output_tokens.extend(sub_tokens)
        return output_tokens

if __name__ == "__main__":
    import torch
    from fantasybert.models.modeling_utils import unilm_mask
    tokenizer = BertTokenizer("S:/bert4torch/resources/chinese_bert_wwm_ext" + "/vocab.txt")
    # tokenizer.add_special_tokens({'additional_special_tokens':['[X]']})

    x=tokenizer('你好呀',text_pair='吃饭',max_length=10, truncation=True,padding=True,return_tensors='pt')
    token_type_ids = x['token_type_ids'].unsqueeze(0)
    print(token_type_ids)
    print(unilm_mask(token_type_ids))

    # pddding  longest first????





#! -*- coding: utf-8 -*-

from setuptools import setup, find_packages

setup(
    name='fantasybert',
    version='0.1.2',
    description='an easy module for bert models',
    long_description='fantasybert: ',
    license='Apache License 2.0',
    url='https://github.com/Changanyue/fantasybert',
    author='Changanyue',
    install_requires=['torch>1.4', 'numpy>=1.17', 'transformers>=4.5.1', 'fitlog'],
    packages=find_packages()
)


import torch
import math
import torch.nn as nn

class Decoder():
    def __init__(self, tokenizer, word2ix, model, device="cuda"):
        super().__init__()
        self.tokenizer = tokenizer
        self.word2ix = word2ix
        self.model = model
        self.device = device
        self.out_max_length = 30
        # output_model_file = os.path.join(args.save_path, "model.bin")
        # self.model.load_state_dict(torch.load(output_model_file, map_location=self.device)['model_state_dict'])

    def generate(self, input_ids, token_type_ids, beam_size=1):
        raise NotImplementedError()


    def beam_search(self, token_ids, token_type_ids, word2ix, beam_size=1):
        sep_id = word2ix["[SEP]"]
        # 用来保存输出序列
        output_ids = [[]]
        # 用来保存累计得分
        output_scores = torch.zeros(token_ids.shape[0], device=self.device)
        for step in range(self.out_max_length):
            scores = self.model.forward(token_ids, token_type_ids)   #  [bs, len, vocab_size]
            if step == 0:
                # 重复beam-size次 输入ids
                token_ids = token_ids.view(1, -1).repeat(beam_size, 1)
                token_type_ids = token_type_ids.view(1, -1).repeat(beam_size, 1)
            ## 计算log 分值 (beam_size, vocab_size)
            logit_score = torch.log_softmax(scores, dim=-1)[:, -1]
            logit_score = output_scores.view(-1, 1) + logit_score  # 累计得分
            ## 取topk的时候我们是展平了然后再去调用topk函数
            # 展平
            logit_score = logit_score.view(-1)
            hype_score, hype_pos = torch.topk(logit_score, beam_size)
            indice1 = hype_pos / scores.shape[-1]  # 行索引
            indice2 = hype_pos % scores.shape[-1]  # 列索引

            # 下面需要更新一下输出了
            new_hype_scores = []
            new_hype_ids = []
            # 为啥有这个[],就是因为要过滤掉结束的序列。
            next_chars = []  # 用来保存新预测出来的一个字符，继续接到输入序列后面，再去预测新字符
            for i_1, i_2, score in zip(indice1, indice2, hype_score):
                i_1 = int(i_1.item())
                i_2 = int(i_2.item())

                hype_id = output_ids[i_1] + [i_2]  # 保存所有输出的序列，而不仅仅是新预测的单个字符

                if i_2 == sep_id:
                    # 说明解码到最后了
                    if score == torch.max(hype_score).item():
                        # 说明找到得分最大的那个序列了 直接返回即可
                        return hype_id[: -1]
                    else:
                        # 完成一个解码了，但这个解码得分并不是最高，因此的话需要跳过这个序列
                        beam_size -= 1
                else:
                    new_hype_ids.append(hype_id)
                    new_hype_scores.append(score)
                    next_chars.append(i_2)  # 收集一下，需要连接到当前的输入序列之后

            output_ids = new_hype_ids

            output_scores = torch.tensor(new_hype_scores, dtype=torch.float32, device=self.device)
            # 现在需要重新构造输入数据了，用上一次输入连接上这次新输出的字符，再输入bert中预测新字符
            token_ids = token_ids[:len(output_ids)].contiguous()  # 截取，因为要过滤掉已经完成预测的序列
            token_type_ids = token_type_ids[: len(output_ids)].contiguous()

            next_chars = torch.tensor(next_chars, dtype=torch.long, device=self.device).view(-1, 1)
            next_token_type_ids = torch.ones_like(next_chars, device=self.device)
            # 连接
            token_ids = torch.cat((token_ids, next_chars), dim=1)
            token_type_ids = torch.cat((token_type_ids, next_token_type_ids), dim=1)
            if beam_size < 1:
                break

        # 如果达到最大长度的话 直接把得分最高的输出序列返回把
        return output_ids[output_scores.argmax().item()]


class AutoRegressiveDecoder(object):
    """通用自回归生成模型解码基类
    包含beam search和random sample两种策略
    """
    def __init__(self, start_id, end_id, maxlen, minlen=1, device='cpu'):
        self.start_id = start_id
        self.end_id = end_id
        self.maxlen = maxlen
        self.minlen = minlen
        self.models = {}
        self.device = device
        if start_id is None:
            self.first_output_ids = torch.empty((1, 0), dtype=int, device=device)
        else:
            self.first_output_ids = torch.tensor([[self.start_id]], device=device)

    @staticmethod
    def wraps(default_rtype='probas', use_states=False):
        """用来进一步完善predict函数
        目前包含: 1. 设置rtype参数，并做相应处理；
                  2. 确定states的使用，并做相应处理；
                  3. 设置温度参数，并做相应处理。
        """
        def actual_decorator(predict):
            def new_predict(
                self,
                inputs,
                output_ids,
                states,
                temperature=1,
                rtype=default_rtype
            ):
                assert rtype in ['probas', 'logits']
                prediction = predict(self, inputs, output_ids, states)

                if not use_states:
                    prediction = (prediction, None)

                if default_rtype == 'logits':
                    prediction = (nn.Softmax(dim=-1)(prediction[0] / temperature), prediction[1])
                elif temperature != 1:
                    probas = torch.power(prediction[0], 1.0 / temperature)
                    probas = probas / probas.sum(axis=-1, keepdims=True)
                    prediction = (probas, prediction[1])

                if rtype == 'probas':
                    return prediction
                else:
                    return torch.log(prediction[0] + 1e-12), prediction[1]

            return new_predict

        return actual_decorator

    # def last_token(self, model):
    #     """创建一个只返回最后一个token输出的新Model
    #     """
    #     if model not in self.models:
    #         outputs = [
    #             keras.layers.Lambda(lambda x: x[:, -1])(output)
    #             for output in model.outputs
    #         ]
    #         self.models[model] = keras.models.Model(model.inputs, outputs)

    #     return self.models[model]

    def predict(self, inputs, output_ids, states=None):
        """用户需自定义递归预测函数
        说明: 定义的时候，需要用wraps方法进行装饰，传入default_rtype和use_states，
             其中default_rtype为字符串logits或probas，probas时返回归一化的概率，
             rtype=logits时则返回softmax前的结果或者概率对数。
        返回: 二元组 (得分或概率, states)
        """
        raise NotImplementedError

    def beam_search(self, inputs_raw, topk, states=None, temperature=1, min_ends=1, add_btz_dim=True):
        """beam search解码
        说明: 这里的topk即beam size；
        返回: 最优解码序列。
        """
        inputs = []
        for i in inputs_raw:
            if isinstance(i, torch.torch.Tensor):
                pass
            elif isinstance(i, (list, tuple, np.ndarray)) and add_btz_dim:
                i = torch.tensor([i], device=self.device)
            elif isinstance(i, (list, tuple, np.ndarray)) and not add_btz_dim:
                i = torch.tensor(i, device=self.device)
            else:
                raise ValueError('Beam search inputs ele only support tensor、array、list、tuple')
            inputs.append(i)

        output_ids, output_scores = self.first_output_ids, torch.zeros(1, device=self.device)
        for step in range(self.maxlen):
            scores, states = self.predict(inputs, output_ids, states, temperature, 'logits')  # 计算当前得分
            if step == 0:  # 第1步预测后将输入重复topk次
                inputs = [i.repeat([topk]+[1]*(len(i.shape)-1)) for i in inputs]
            scores = output_scores.reshape((-1, 1)) + scores  # 综合累积得分
            indices = scores.flatten().argsort(dim=-1, descending=True)[:topk]  # 仅保留topk
            indices_1 = torch.div(indices, scores.shape[1], rounding_mode='trunc')  # 行索引
            indices_2 = (indices % scores.shape[1]).reshape((-1, 1))  # 列索引
            output_ids = torch.cat([output_ids[indices_1], indices_2], 1)  # 更新输出
            output_scores = torch.take_along_dim(scores, indices, dim=None)  # 更新得分
            is_end = output_ids[:, -1] == self.end_id  # 标记是否以end标记结束
            end_counts = (output_ids == self.end_id).sum(1)  # 统计出现的end标记
            if output_ids.shape[1] >= self.minlen:  # 最短长度判断
                best = output_scores.argmax()  # 得分最大的那个
                if is_end[best] and end_counts[best] >= min_ends:  # 如果已经终止
                    return output_ids[best]  # 直接输出
                else:  # 否则，只保留未完成部分
                    flag = ~is_end | (end_counts < min_ends)  # 标记未完成序列
                    if not flag.all():  # 如果有已完成的
                        inputs = [i[flag] for i in inputs]  # 扔掉已完成序列
                        output_ids = output_ids[flag]  # 扔掉已完成序列
                        output_scores = output_scores[flag]  # 扔掉已完成序列
                        end_counts = end_counts[flag]  # 扔掉已完成end计数
                        topk = flag.sum()  # topk相应变化
        # 达到长度直接输出
        return output_ids[output_scores.argmax()]

    def random_sample(
        self,
        inputs,
        n,
        topk=None,
        topp=None,
        states=None,
        temperature=1,
        min_ends=1
    ):
        """随机采样n个结果
        说明: 非None的topk表示每一步只从概率最高的topk个中采样；而非None的topp
             表示每一步只从概率最高的且概率之和刚好达到topp的若干个token中采样。
        返回: n个解码序列组成的list。
        """
        inputs = [torch.tensor([i], device=self.device) for i in inputs]
        output_ids = self.first_output_ids
        results = []
        for step in range(self.maxlen):
            probas, states = self.predict(inputs, output_ids, states, temperature, 'probas')  # 计算当前概率
            probas /= probas.sum(dim=-1, keepdims=True)  # 确保归一化
            if step == 0:  # 第1步预测后将结果重复n次
                probas = probas.repeat([n]+[1]*(len(probas.shape)-1))
                inputs = [i.repeat([n]+[1]*(len(i.shape)-1)) for i in inputs]
                output_ids = output_ids.repeat([n]+[1]*(len(output_ids.shape)-1))
            if topk is not None:
                k_indices = probas.argsort(dim=-1, descending=True)[:, :topk]  # 仅保留topk
                probas = torch.take_along_dim(probas, k_indices, dim=1)  # topk概率
                probas /= probas.sum(dim=1, keepdims=True)  # 重新归一化
            if topp is not None:
                p_indices = probas.argsort(dim=-1, descending=True)  # 从高到低排序
                probas = torch.take_along_dim(probas, p_indices, dim=-1)  # 排序概率
                cumsum_probas = torch.cumsum(probas, dim=-1)  # 累积概率
                flag = torch.roll(cumsum_probas >= topp, 1, dims=1)  # 标记超过topp的部分
                flag[:, 0] = False  # 结合上面的torch.roll，实现平移一位的效果
                probas[flag] = 0  # 后面的全部置零
                probas /= probas.sum(dim=1, keepdims=True)  # 重新归一化

            sample_func = lambda p: torch.multinomial(p, 1)  # 按概率采样函数
            sample_ids = torch.stack([sample_func(p) for p in probas])
            sample_ids = sample_ids.reshape((-1, 1))  # 对齐形状
            if topp is not None:
                sample_ids = torch.take_along_dim(p_indices, sample_ids, dim=1)  # 对齐原id
            if topk is not None:
                sample_ids = torch.take_along_dim(k_indices, sample_ids, dim=1)  # 对齐原id
            output_ids = torch.cat([output_ids, sample_ids], 1)  # 更新输出
            is_end = output_ids[:, -1] == self.end_id  # 标记是否以end标记结束
            end_counts = (output_ids == self.end_id).sum(1)  # 统计出现的end标记
            if output_ids.shape[1] >= self.minlen:  # 最短长度判断
                flag = is_end & (end_counts >= min_ends)  # 标记已完成序列
                if flag.any():  # 如果有已完成的
                    for ids in output_ids[flag]:  # 存好已完成序列
                        results.append(ids)
                    flag = (flag == False)  # 标记未完成序列
                    inputs = [i[flag] for i in inputs]  # 只保留未完成部分输入
                    output_ids = output_ids[flag]  # 只保留未完成部分候选集
                    end_counts = end_counts[flag]  # 只保留未完成部分end计数
                    if len(output_ids) == 0:
                        break
        # 如果还有未完成序列，直接放入结果
        for ids in output_ids:
            results.append(ids)
        # 返回结果
        return results


def Bertpooling(bert_outputs, pool_type='pooler'):
    encoded_layers = bert_outputs.hidden_states
    sequence_output = bert_outputs.last_hidden_state

    if pool_type == 'pooler':
        pooled_output = bert_outputs.pooler_output

    elif pool_type == 'first-last-avg':
        first = encoded_layers[1]
        last = encoded_layers[-1]
        seq_length = first.size(1)

        first_avg = torch.avg_pool1d(first.transpose(1, 2), kernel_size=seq_length).squeeze(-1)  # batch, hid_size
        last_avg = torch.avg_pool1d(last.transpose(1, 2), kernel_size=seq_length).squeeze(-1)  # batch, hid_size
        pooled_output = torch.avg_pool1d(
            torch.cat([first_avg.unsqueeze(1), last_avg.unsqueeze(1)], dim=1).transpose(1, 2),
            kernel_size=2).squeeze(-1)

    elif pool_type == 'last-avg':
        pooled_output = torch.mean(sequence_output, 1)

    elif pool_type == 'cls':
        pooled_output = sequence_output[:, 0]

    else:
        raise TypeError('Please the right pool_type from cls, pooler, first-last-avg and last-avg !!!')

    return pooled_output


def unilm_mask(token_type_ids):
    seq_len = token_type_ids.size(1)
    bs = token_type_ids.size(0)
    idxs = torch.cumsum(token_type_ids, dim=1)
    mask = (idxs.unsqueeze(1) <= idxs.unsqueeze(2)).long()

    last_one_idx = 0
    for bsz in range(bs):
        for i, v in enumerate(token_type_ids[bsz]):
            if v == 1:
                last_one_idx = i
        pad_mask = torch.tensor([1]*(last_one_idx+1) + [0]* (seq_len-last_one_idx-1))
        pad_mask = (pad_mask.unsqueeze(0) * pad_mask.unsqueeze(1)).to(mask.device)
        mask[bsz] = mask[bsz]*pad_mask
        last_one_idx = 0

    return mask


def gelu(x):
    """
    """
    return x * 0.5 * (1.0 + torch.erf(x / math.sqrt(2.0)))


ACT2FN = {"gelu": gelu, "relu": torch.nn.functional.relu}

class BertPredictionHeadTransform(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        self.transform_act_fn = ACT2FN[config.hidden_act]
        self.LayerNorm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)

    def forward(self, hidden_states):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.transform_act_fn(hidden_states)
        hidden_states = self.LayerNorm(hidden_states)
        return hidden_states


class BertLMPredictionHead(nn.Module):
    def __init__(self, config, bert_model_embedding_weights):
        super().__init__()
        self.transform = BertPredictionHeadTransform(config)

        # The output weights are the same as the input embeddings, but there is
        # an output-only bias for each token.
        self.decoder = nn.Linear(config.hidden_size, config.vocab_size, bias=False)

        self.decoder.weight = bert_model_embedding_weights
        self.bias = nn.Parameter(torch.zeros(config.vocab_size))

        # Need a link between the two variables so that the bias is correctly resized with `resize_token_embeddings`
        self.decoder.bias = self.bias

    def forward(self, hidden_states):
        hidden_states = self.transform(hidden_states)
        hidden_states = self.decoder(hidden_states)
        return hidden_states


def build_bert_optim_parameters(model, weight_decay):
    no_decay = ['bias', 'LayerNorm.weight']
    optimizer_grouped_parameters = [
        {
            "params": [p for n, p in model.named_parameters() if not any(nd in n for nd in no_decay)],
            "weight_decay": weight_decay,
        },
        {
            "params": [p for n, p in model.named_parameters() if any(nd in n for nd in no_decay)],
            "weight_decay": 0.0
        },
    ]
    return optimizer_grouped_parameters



#! -*- coding: utf-8 -*-

"""
Logger 是记录日志的模块，logger封装了logging模块的Logger，
具体使用方式与直接使用logging.Logger相同，同时也新增一些简单好用的API

# logger 新增的API
# 将日志输出到文件，以及输出的日志等级
logger.add_file('/path/to/log', level='INFO')
# 定义在命令行中的显示格式和日志等级
logger.set_stdout('tqdm', level='WARN')
"""



__all__ = [
    'logger',
    'init_logger_dist'
]

import logging
import logging.config
import os
import sys
import warnings
from torch import distributed as dist

ROOT_NAME = 'FantasyBert'

try:
    import fitlog
except ImportError:
    fitlog = None
try:
    from tqdm.auto import tqdm
except ImportError:
    tqdm = None

if tqdm is not None:
    class TqdmLoggingHandler(logging.Handler):
        def __init__(self, level=logging.INFO):
            super().__init__(level)

        def emit(self, record):
            try:
                msg = self.format(record)
                tqdm.write(msg)
                self.flush()
            except (KeyboardInterrupt, SystemExit):
                raise
            except:
                self.handleError(record)
else:
    class TqdmLoggingHandler(logging.StreamHandler):
        def __init__(self, level=logging.INFO):
            super().__init__(sys.stdout)
            self.setLevel(level)


def _get_level(level):
    if isinstance(level, int):
        pass
    else:
        level = level.lower()
        level = {'info': logging.INFO, 'debug': logging.DEBUG,
                 'warn': logging.WARN, 'warning': logging.WARN,
                 'error': logging.ERROR}[level]
    return level


def _add_file_handler(logger, path, level='INFO'):
    for h in logger.handlers:
        if isinstance(h, logging.FileHandler):
            if os.path.abspath(path) == h.baseFilename:
                # file path already added
                return

    # File Handler
    if os.path.exists(path):
        assert os.path.isfile(path)
        warnings.warn('log already exists in {}'.format(path))
    dirname = os.path.abspath(os.path.dirname(path))
    os.makedirs(dirname, exist_ok=True)

    file_handler = logging.FileHandler(path, mode='a')
    file_handler.setLevel(_get_level(level))
    file_formatter = logging.Formatter(fmt='%(asctime)s - %(module)s - [%(levelname)s] - %(message)s',
                                       datefmt='%Y/%m/%d %H:%M:%S')
    file_handler.setFormatter(file_formatter)
    logger.addHandler(file_handler)


def _set_stdout_handler(logger, stdout='tqdm', level='INFO'):
    level = _get_level(level)
    if stdout not in ['none', 'plain', 'tqdm']:
        raise ValueError('stdout must in one of {}'.format(['none', 'plain', 'tqdm']))
    # make sure to initialize logger only once
    stream_handler = None
    for i, h in enumerate(logger.handlers):
        if isinstance(h, (logging.StreamHandler, TqdmLoggingHandler)):
            stream_handler = h
            break
    if stream_handler is not None:
        logger.removeHandler(stream_handler)

    # Stream Handler
    if stdout == 'plain':
        stream_handler = logging.StreamHandler(sys.stdout)
    elif stdout == 'tqdm':
        stream_handler = TqdmLoggingHandler(level)
    else:
        stream_handler = None

    if stream_handler is not None:
        stream_formatter = logging.Formatter('%(message)s')
        stream_handler.setLevel(level)
        stream_handler.setFormatter(stream_formatter)
        logger.addHandler(stream_handler)


class FastNLPLogger(logging.getLoggerClass()):
    def __init__(self, name):
        super().__init__(name)

    def add_file(self, path='./log.txt', level='INFO'):
        r"""add log output file and the output level"""
        _add_file_handler(self, path, level)

    def set_stdout(self, stdout='tqdm', level='INFO'):
        r"""set stdout format and the output level"""
        _set_stdout_handler(self, stdout, level)


logging.setLoggerClass(FastNLPLogger)


# print(logging.getLoggerClass())
# print(logging.getLogger())

def _init_logger(path=None, stdout='tqdm', level='INFO'):
    r"""initialize logger"""
    level = _get_level(level)

    # logger = logging.getLogger()
    logger = logging.getLogger(ROOT_NAME)
    logger.propagate = False
    logger.setLevel(1)  # make the logger the lowest level

    _set_stdout_handler(logger, stdout, level)

    # File Handler
    if path is not None:
        _add_file_handler(logger, path, level)

    return logger


def _get_logger(name=None, level='INFO'):
    level = _get_level(level)
    if name is None:
        name = ROOT_NAME
    assert isinstance(name, str)
    if not name.startswith(ROOT_NAME):
        name = '{}.{}'.format(ROOT_NAME, name)
    logger = logging.getLogger(name)
    logger.setLevel(level)
    return logger


logger = _init_logger(path=None, level='INFO')


def init_logger_dist():
    global logger
    rank = dist.get_rank()
    logger.setLevel(logging.INFO if rank == 0 else logging.WARNING)
